# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 27 Jan 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.args` module.
"""


import pytest

from ..args import SmartHelpFormatter


class TestSmartHelpFormatter:
    def test_get_indent(self):
        class NoLeading(SmartHelpFormatter):
            KEEP_LEADING = False
        fmt = NoLeading('', 2, 24, 80)
        assert fmt._get_indent('x') == 0
        assert fmt._get_indent('    x') == 0
        assert fmt._get_indent(' \t  x') == 0
        assert fmt._get_indent(' \t  x\t\t') == 0

        fmt = SmartHelpFormatter('', 2, 24, 80)
        assert fmt._get_indent('x') == 0
        assert fmt._get_indent('    x') == 4
        assert fmt._get_indent(' \t  x') == 7
        assert fmt._get_indent(' \t  x\t\t') == 7

    def test_split_lines(self):
        class NoNewlines(SmartHelpFormatter):
            KEEP_NEWLINES = False
        fmt = NoNewlines('', 2, 24, 80)
        assert fmt._split_lines('x', 80) == ['x']
        assert fmt._split_lines('    x', 80) == ['    x']
        assert fmt._split_lines(' \t  x', 80) == ['       x']
        assert fmt._split_lines('x\n', 80) == ['x']
        assert fmt._split_lines('    x\ny', 80) == ['    x y']
        assert fmt._split_lines(' \t  x\ny\n      zzz zzz', 10) == ['       x y', '       zzz', '       zzz']

        fmt = SmartHelpFormatter('', 2, 24, 80)
        assert fmt._split_lines('x', 80) == ['x']
        assert fmt._split_lines('    x', 80) == ['    x']
        assert fmt._split_lines(' \t  x', 80) == ['       x']
        assert fmt._split_lines('x\n', 80) == ['x']
        assert fmt._split_lines('    x\ny', 80) == ['    x', 'y']
        assert fmt._split_lines(' \t  x\ny\n      zzz zzz', 10) == ['       x', 'y', '      zzz', '      zzz']

    def test_fill_text(self):
        class NoNewlines(SmartHelpFormatter):
            KEEP_NEWLINES = False
        fmt = NoNewlines('', 2, 24, 80)
        assert fmt._fill_text('x', 80, '  ') == '  x'
        assert fmt._fill_text('    x', 80, '  ') == '      x'
        assert fmt._fill_text(' \t  x', 80, '  ') == '         x'
        assert fmt._fill_text('x\n', 80, '  ') == '  x'
        assert fmt._fill_text('    x\ny', 80, '  ') == '      x y'
        assert fmt._fill_text(' \t  x\ny\n      zzz zzz', 12, '  ') == '         x y\n         zzz\n         zzz'

        fmt = SmartHelpFormatter('', 2, 24, 80)
        assert fmt._fill_text('x', 80, '  ') == '  x'
        assert fmt._fill_text('    x', 80, '  ') == '      x'
        assert fmt._fill_text(' \t  x', 80, '  ') == '         x'
        assert fmt._fill_text('x\n', 80, '  ') == '  x'
        assert fmt._fill_text('    x\ny', 80, '  ') == '      x\n  y'
        assert fmt._fill_text(' \t  x\ny\n      zzz zzz', 12, '  ') == '         x\n  y\n        zzz\n        zzz'



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 12 Jan 2022: Initial Coding

"""
Global configuration and fixture setup for pytest.

This configuration depends on the :mod:`haggis.tests.options` plugin.
"""


from errno import EEXIST
from os import makedirs
from os.path import join
from warnings import warn

from pytest import fixture


@fixture(scope='session')
def plots(request):
    """
    Enables debugging for the fixtures/tests that care about it.

    This fixture will only be set to `True` if the `--plots`
    command-line option is set through the :mod:`skg.tests.options`
    plugin.
    """
    if request.config.getoption('--plots', False):
        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot
        except ImportError:
            warn('Unable to import matplotlib: '
                 'plotting will not be enabled.')
            return False

        try:
            from .util import FOLDER
            makedirs(join(request.config.rootdir, FOLDER))
        except OSError as e:
            if e.errno != EEXIST:
                raise
        return True

    return False



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 09 Jan 2021: Initial Coding

"""
Recipes specifically for manipulating and emulating mappings, namespaces
and the like.

Implementations based on some of the recipes provided in the Python
documentation, and other sources like Stack Overflow.
"""

__all__ = [
    'dict_merge', 'dict_select', 'mapping_context', 'Namespace',
    'option_lookup', 'RecursiveDict', 'setdefaults',
]


from collections.abc import Mapping
from itertools import starmap
from warnings import warn

from .exceptions import ErrorTransform
from .recipes import consume


def dict_select(dic, keys=None, exclude=None, extra='ignore'):
    """
    Filter a dictionary so only the specified keys are present.

    Parameters
    ----------
    dic : dict
        The dictionarty to filter.
    keys : iterable or None
        The keys to include in the output. Another dictionary may be
        used since it iterates over its keys by default. :py:obj:`None`
        means to use all keys.
    exclude : container or None
        The keys to exclude. Anything that supports the `in` operator is
        valid here. `exclude` takes precedence over `keys`: no keys in
        `exclude` will be present in the output, even if they are
        present in `keys`.
    extra : {'ignore', 'err', 'warn'}
        How to handle members of `dic` that are neither in `keys` nor
        explicitly listed in `exclude`:

        ignore :
            Skip over extra keys.
        err :
            Issue an error if invalid keys are found.
        warn :
            Issue a warning if extra keys are found.

        Values other than `'ignore'` will compare sets of keys.

    Return
    ------
    selection : dict
        A new `dict` object, even if `keys` is a superset of the actual
        keys found in `dic`.

    Notes
    -----
    The default behavior is just to make a copy of `dic`.
    """
    input_keys = set(dic.keys())
    select_keys = input_keys if keys is None else set(keys)
    exclude_keys = set() if exclude is None else set(exclude)

    if extra != 'ignore':
        extra_keys = (input_keys - select_keys) - exclude_keys
        extra_fmt = ', '.join(str(k) for k in extra_keys)
        message = 'Found extra keys: {}'.format(extra_fmt)
        if extra == 'warn':
            warn(message)
        elif extra == 'err':
            raise KeyError(message)
        else:
            raise ValueError('Invalid value of `extra`: "{}"'.format(extra))

    selection = (input_keys & select_keys) - exclude_keys

    return {k: dic[k] for k in selection}


def dict_merge(parent, child, keys=None, exclude=None, key=None):
    """
    Filter a parent dictionary and override it with values from a
    child, if supplied as a mapping.

    Parameters
    ----------
    parent : dict
        The base mapping to get the keys from.
    child : dict, value or None
        If a mapping type, `child[key]` will be the main value and the
        remaining keys will override the values obtained from `parent`.
        Otherwise, `child` will be the main value. If :py:obj:`None`,
        treated as missing entirely.
    keys : iterable or None
        An iterable of the keys to extract from `parent` and `child`, if
        it is a mapping. Should not contain `key`. :py:obj:`None` means
        to use all the keys of `parent`.
    exclude : container or None
        Keys to omit from the final result, even if they are present in
        `keys`.
    key :
        The name of the main value in `parent`, if it is not overriden
        by `child`. If `child` is a mapping, either it or `parent` must
        contain this key. Otherwise, the value of child is the value of
        the key.

    Return
    ------
    main: value (opt)
        The value of the main key, only returned if `key` is not
        :py:obj:`None`. If `child` is a mapping type, either it or
        `parent` will provide the key. Otherwise, it will be `child`
        itself.
    selection:
        The selected values of `parent`, possibly overriden by `child`
        if it is a mapping.
    """
    selection = dict_select(parent, keys, exclude)
    cmap = isinstance(child, Mapping)

    if cmap:
        selection.update(dict_select(child, keys, exclude))

    if key is None:
        return selection

    if child is None:
        value = parent[key]
    elif cmap:
        value = child[key] if key in child else parent[key]
    else:
        value = child

    return value, selection


def setdefaults(mapping, *args, **kwargs):
    """
    Update missing keys in this mapping based on supplied iterables and
    mappings.

    This is similar to :py:meth:`dict.update`, except that only missing
    keys are added.

    Parameters
    ----------
    mapping :
        The dictionary to update. If the type does not have a
        `setdefault` method, the udpate will default to using
        `__contains__` and `__setitem__` directly.
    *args :
        Each positional arguments may be a mapping or an iterable of
        two-element iterables. Iterables are applied in order. Only the
        first instance of a duplicated key is ever considered.
    **kwargs :
        Any additional keywords to insert. These are applied after the
        iterables, if any.
    """
    # Find suitable setdefault implementation
    if hasattr(mapping, 'setdefault') and callable(mapping.setdefault):
        func = mapping.setdefault
    else:
        def func(key, default):
            if not key in mapping:
                ret = mapping[key] = default
            else:
                ret = mapping[key]
            # For correctness
            return ret

    for arg in args:
        if isinstance(arg, Mapping):
            arg = arg.items()
        consume(starmap(func, arg))
    consume(starmap(func, kwargs.items()))


class RecursiveDict(dict):
    """
    Mapping that allows recursive lookup of keys.

    Lookup can be controlled by specifying a key type and/or a value
    type. A value suitable for recursion that is not found as a key
    raises a :py:exc:`KeyError`, while a value of the wrong type raises
    a :py:exc:`TypeError`. Requesting a key of the wrong type raises a
    :py:exc:`TypeError` as well.

    Only lookup operations are different from :py:class:`dict`\ :
    :py:meth:`__getitem__`, :py:meth:`setdefault` and :py:meth:`get`.
    :py:meth:`~dict.pop` and :py:meth:`~dict.popitem` are not modified.
    Full control is provided through the unaltered setter operations.

    Here is an example of how :py:attr:`value_type` affects the lookup::

        d = RecursiveDict({'a': 1, 'b': 'a'})
        print(d.setdefault('d', 'c'))  # Prints "c"

    vs::

        d = RecursiveDict({'a': 1, 'b': 'a'}, value_type=int)
        print(d.setdefault('d', 'c'))  # Raises "KeyError: 'c'"

    .. py:attribute:: key_type

       A class object indicating the type that may be used as a key.
       :py:obj:`None` indicates "any". The default is :py:class:`str`.
       Values of this type indicate recursive lookup.

    .. py:attribute:: value_type

       A class object indicating the type that may be used as a value.
       :py:obj:`None`, the default, indicates "any" type.
    """
    __slots__ = ('key_type', 'value_type')

    def __init__(self, *args, key_type=str, value_type=None, **kwargs):
        """
        Construct a new mapping with the specified key and value types.

        All normal :py:class:`dict` constructor arguments are accepted.

        `key_type` and `value_type` may be classes, tuples of classes,
        or :py:obj:`None`. The latter case is equivalent to "anything
        goes". The dictionary will accept keys and values of the wrong
        type, but will raise an error on lookup, so it is best to leave
        either `key_type` or `value_type` as :py:obj:`None`.

        If both types are constrained and there is overlap between them,
        `key_type` always takes precedence: a value that can be a
        `key_type` is always looked up as a key. Only when the lookup
        fails is the value returned.

        If `key_type` is constrained, but `value_type` is not, lookup
        continues until a :py:exc:`KeyError` is raised or a
        non-\ `key_type` value is encountered. Similarly, if only
        `value_type` is constrained, lookup continues until a
        :py:exc:`KeyError` is raised or a value of the requested type
        is found.

        There is no constraint on either (both are `None`), recursion
        will stop only when a value is not present in the dictionary as
        a key.
        """
        self.key_type = key_type
        self.value_type = value_type
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        """
        Recursively fetch an item from this dictionary.
        """
        return self._conditional_get(key, iskey=True)[1]

    def setdefault(self, key, default=None):
        """
        Set the key to the specified value if not found, return the
        value for the key.

        The default value is set exactly as specified, but the return
        value is dereferenced.
        """
        if key in self:
            return self[key]
        self[key] = default
        return self._conditional_get(default)[1]

    def get(self, key, default=None):
        """
        Retrieve a fully dereferenced value for `key` if present, or
        `default` if not.

        `default` gets dereferenced if the key is not present.
        """
        if key in self:
            return self[key]
        return self._conditional_get(default)[1]

    def final_key(self, key):
        """
        In a recursive reference, retreive the final key that actually
        contains the value mapping.

        For example, given::

            x = RecursiveDict(key_type=str, value_type=int)
            x['a'] = 'b'
            x['b'] = 'c'
            x['c'] = 1

        The result of :py:meth:`final_key` on any of the keys defined
        above would be ``'c'``.

        If the final key does not contain a value, it will still be
        returned. This means that if the dictionary above were to have
        ``x['c'] = 'd'`` (a broken recursion), :py:meth:`final_key`
        would return ``'c'`` regardless.

        Passing in a missing key will raise a :py:exc:`KeyError` as
        usual.
        """
        return self._conditional_get(key, iskey=True, err=False)[0]

    def _conditional_get(self, obj, iskey=False, err=True):
        """
        Check if the object is a possible key, and correctly retrieve
        the dereferenced value for it.

        If the value is a key type, but is not a valid key, it must also
        be a valid value type to avoid raising a :py:exc:`KeyError`.

        Parameters
        ----------
        obj :
            The object to search for.
        iskey : bool
            :py:obj:`True` if `obj` is required to be a key on the
            first iteration. Otherwise it represents a value (which may
            be a recursive key).
        err : bool
            Whether or not to raise an error if the key is not found.
        """
        key = obj if iskey else None
        while self._check_key_or_value_type(obj, iskey=iskey):
            try:
                new = super().__getitem__(obj)
            except KeyError:
                if err and (iskey or
                            not self._check_value_type(obj, err=False)):
                    raise
                break
            else:
                key = obj
                obj = new
            iskey = False
        return key, obj

    def _check_key_type(self, key, err=True):
        """
        Verify that the selected object is a valid key.

        Return the status, or raise a :py:exc:`TypeError` if `err` is
        :py:obj:`True`.
        """
        if self.key_type is None or isinstance(key, self.key_type):
            return True
        if err:
            raise TypeError('{} not allowed for keys'.format(type(key)))
        return False

    def _check_value_type(self, value, err=True):
        """
        Verify that the selected object is a valid value.

        Return the status, or raise a :py:exc:`TypeError` if `err` is
        :py:obj:`True`.
        """
        if self.value_type is None or isinstance(value, self.value_type):
            return True
        if err:
            raise TypeError('{} not allowed in mapping'.format(type(value)))
        return False

    def _check_key_or_value_type(self, obj, iskey):
        """
        Check if the object is of a proper key or value type, depending
        on the iteration count.

        If `iskey` is :py:obj:`True`, the object must be a key. All
        subsequent iterations return :py:obj:`True` if the object is a
        key, :py:obj:`False` if it is a valid value, and raise an error
        if it is neither.
        """
        check = self._check_key_type(obj, err=iskey)
        if check:
            return True
        return not self._check_value_type(obj, err=True)


def option_lookup(name, mapping, option, key_func=None, value_func=None,
                  err_type=ValueError, key_err=KeyError):
    """
    Perform a lookup of an option in a mapping whose keys represent
    the valid options.

    If an invalid option is selected, it an `err_type` error is raised
    (:py:exc:`ValueError` by default).

    Parameters
    ----------
    name : str
        The name of the mapping to display if an error occurs. If the
        option is a function argument, the name of the argument is a
        good name choice.
    mapping : mapping
        The mapping to perform a the lookup in.
    option : object
        The key to look up in mapping.
    key_func : callable or None
        The transformation to apply to `option` to obtain the actual
        key to look up. If :py:obj:`None` (the default), no
        transformation is done. A common choice for this argument is
        :py:meth:`str.casefold` or :py:meth:`str.lower`, for
        case-insensitive string keys.
    value_func : callable or None
        The transformation to apply to the value that is found. If
        :py:obj:`None` (the default), no transformation is done. If
        supplied, must be a callable that accepts three arguments:

        - ``option``
        - ``key``: ``option`` transformed by ``key_func``
        - ``value``: The raw value retreived from the dictionary.

    err_type : type
        The type of the :py:exc:`Exception` to raise. Default is
        :py:exc:`ValueError`.
    key_err : type or tuple[type]
        The type of error expected from a failed lookup in `mapping`.
        If multiple types are to be expected, a :py:class:`tuple` of
        types may be supplied. Defaults to :py:exc:`KeyError`.
    """
    with ErrorTransform(key_err, err_type, 'Invalid value for `{0}`: {1!r}',
                        name, option):
        key = option if key_func is None else key_func(option)
        value = mapping[key]
    # This is not in the exception handler because its your own fault:
    # Option can be anything since its user supplied, but not mapping values
    if value_func is not None:
        value = value_func(option, key, value)
    return value


class mapping_context:
    """
    A context manager for temporarily modifying the keys of a
    mapping.

    The context manager has a :py:meth:`delete` method that allows
    values to be removed as well. The method accepts any number of
    attribute names to delete and returns the context manager, so it can
    be used in a :ref:`with` block directly. Missing key names are
    ignored.

    All mapping values are reset to their original values when the
    manager exits. Nesting multiple instances of this context manager
    has a cumulative effect.

    The context manager is reentrant: the manager can be entered and
    exited and modified before re-entry as many times as necessary.

    Sample usage::

        d = {'a': 1, 'b': 2}
        print(d)
        with mapping_context(d, b=0, c=3):
            print(d)
            with mapping_context(d, b=1).delete('a'):
                print(d)
            print(d)
        print(d)

    Will result in::

        {'a': 1, 'b': 2}
        {'a': 1, 'b': 0, 'c': 3}
        {'b': 1, 'c': 3}
        {'b': 0, 'c': 3, 'a': 1}
        {'b': 2, 'a': 1}

    Instances have three documented attributes:

    .. py:attribute:: mapping

       The :py:class:`dict`\ -like object that this context manager
       applies to. The entire interface to the mapping is through the
       :py:meth:`getfunc` and :py:meth:`setfunc` methods, so that a
       context manager for non-mapping types can be simulated by
       subclassing.

    .. py:attribute:: updates

       A mapping of the added keys to their prior values. This mapping
       is unordered.

    .. py:attribute:: sentinel

       A special marker object guaranteed not to exists in the original
       mapping, used to mark deleted keys.
    """
    __slots__ = ('mapping', 'updates', 'sentinel')

    def __init__(self, mapping, *args, **kwargs):
        """
        Initialize a new context manager to replace the
        specified elements of the given mapping.

        Positional arguments represent mappings that contain keys and
        values to replace or add, each one in a format that would be
        accepted by :py:meth:`dict.update`.

        Keyword arguments are individual keys to update.
        """
        self.sentinel = object()
        self._init(mapping, *args, **kwargs)

    def __enter__(self):
        """
        Enter the context manager by updating the requested
        elements, return the context manager itself.
        """
        for key, value in self.updates.items():
            self.updates[key] = self.getfunc(key)
            self.setfunc(key, value)
        return self

    def __exit__(self, *args, **kwargs):
        """
        Restore the elements of the mapping.
        """
        # Convert to list to avoid modification during iteration
        for item in self.updates.items():
            self.setfunc(*item)

    def delete(self, *keys):
        """
        Remove all the keys named in ``keys`` from the mapping, but keep
        a record of them to be restored on exit.

        Return the context manager itself.
        """
        for key in keys:
            self.updates[key] = self.sentinel
        return self

    def chain(self, *args, **kwargs):
        """
        Create a chained :py:class:`mapping_context` with the same
        mapping class and :py:attr:`sentinel` object as this mapping.

        This method allows the example in the class docs to be rewritten
        as::

            d = {'a': 1, 'b': 2}
            print(d)
            with mapping_context(d, b=0, c=3) as mc:
                print(d)
                with mc.chain(b=1).delete('a'):
                    print(d)
                print(d)
            print(d)

        Returns a new :py:class:`mapping_context`.
        """
        cls = type(self)
        dc = cls.__new__(cls)
        dc.sentinel = self.sentinel
        dc._init(self.mapping, *args, **kwargs)
        return dc

    def _init(self, mapping, *args, **kwargs):
        """
        Sets ``mapping`` and records the modified and added keys
        similarly to :py:meth:`dict.update`, except that it allows
        multiple positional arguments, which may be of mixed types.
        Arguments are added in the following order:

        1. ``args`` is processed in order. Each element is added in the
           order of its iterator.
        2. ``kwargs`` is added in the order of its iterator.

        Later duplicate keys silently override earlier ones.
        """
        self.mapping = mapping
        self.updates = {}
        for arg in args:
            self.updates.update(arg)
        self.updates.update(kwargs)

    def getfunc(self, key):
        """
        A customizable function to get a single element of the mapping.

        If `key` is not present, this function must return
        :py:attr:`sentinel`.

        The default implementation works for most Python builtin mapping
        types that support a :py:meth:`~dict.get` method.
        """
        return self.mapping.get(key, self.sentinel)

    def setfunc(self, key, value):
        """
        A customizable function to set a single element of the mapping.

        If value is :py:attr:`sentinel`, the element should be deleted.

        The default implementation works for most Python builtin mapping
        types that support :py:meth:`~object.__delitem__` and
        :py:meth:`~object.__setitem__` methods.
        """
        if value is self.sentinel:
            del self.mapping[key]
        else:
            self.mapping[key] = value


class object_context(mapping_context):
    """
    An extension of :py:class:`mapping_context` that is tailored for
    mutable objects that support :py:func:`getattr`, :py:func:`setattr`
    and :py:func:`delattr`.
    """
    __slots__ = ()

    def getfunc(self, key):
        """
        Mapping getter customized for :py:class:`object` instead of
        :py:class:`dict`.
        """
        return getattr(self.mapping, key, self.sentinel)

    def setfunc(self, key, value):
        """
        Mapping setter and deleter customized for :py:class:`object`
        instead of :py:class:`dict`.
        """
        if value is self.sentinel:
            delattr(self.mapping, key)
        else:
            setattr(self.mapping, key, value)


class _NamespaceOverrideContext(object_context):
    """
    An extension of :py:class:`object_context` designed specifically for
    :meth:`Namespace.override`.

    The inputs are restricted to keywords only, and the managed object
    is returned instead of the context manager.
    """
    __slots__ = ()

    def __init__(self, namespace, **kwargs):
        """
        Initialize a new context manager to replace the specified
        elements of the given namespace.
        """
        super().__init__(namespace, **kwargs)

    def __enter__(self):
        """
        Enter the context manager and return the mapping rather than
        `self`.
        """
        return super().__enter__().mapping


class Namespace:
    """
    A simple namespace object.

    The class is mutable. It implements containment checks. It can be
    converted to a dictionary using :py:func:`vars`. That being said, it
    supports a dictionary-like interface for elements whose names are not
    valid python identifiers, or are shadowed by descriptors.

    This class originated with :py:class:`argparse.Namespace` and
    :py:class:`types.SimpleNamespace`.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new namespace with the specified named arguments.

        Key-value pairs in each iterable of `args` are added in order,
        followed by the mapping `kwargs`.
        """
        for arg in args:
            self.__dict__.update(arg)
        self.__dict__.update(kwargs)

    def __contains__(self, name):
        """
        Check if the namespace contains `name`.
        """
        return name in self.__dict__

    def __eq__(self, other):
        """
        Check if this object is equal to another namespace.

        Two namespaces are considered equal if they are both subclasses
        of :py:class:`Namespace` and contain the same mappings.
        """
        if not isinstance(other, __class__):
            return NotImplemented
        return self.__dict__ == other.__dict__

    def __repr__(self):
        """
        Returns a usable string representation of this namespace.
        """
        return '{}({!r})'.format(type(self).__name__, self.__dict__)

    def __getitem__(self, key):
        """
        Retrieve an item directly from the dictionary.

        Useful for items whose names are not valid python identifiers.

        Parameters
        ----------
        key :
            The key of the object to retreive. Does not have to be a
            string.

        Return
        ------
        object :
            The value mapped to the specified key. Raises an error if
            the key is not found.
        """
        return self.__dict__[key]

    def __setitem__(self, key, value):
        """
        Set an item directly into the dictionary.

        Useful for items whose names are not valid python identifiers.

        Parameters
        ----------
        key :
            The key of the object to retreive. Does not have to be a
            string.
        value :
            The value mapped to the specified key.
        """
        self.__dict__[key] = value

    def __delitem__(self, key):
        """
        Delete a mapping directly from the dictionary.

        Useful for items whose names are not valid python identifiers.

        Parameters
        ----------
        key :
            The key of the mapping to delete. Does not have to be a
            string.
        """
        del self.__dict__[key]

    def __len__(self):
        """
        The number of items in this namespace.
        """
        return len(self.__dict__)

    def get(self, key, default=None):
        """
        :py:class:`dict`-like :py:meth:`~dict.get` operation on the
        namespace's mapping.
        """
        return self.__dict__.get(key, default)

    def setdefault(self, key, value):
        """
        :py:class:`dict`-like :py:meth:`~dict.setdefault` operation on
        the namespace's mapping.
        """
        return self.__dict__.setdefault(key, value)

    def items(self):
        """
        An iterator over the items in this namepace's mapping.
        """
        return self.__dict__.items()

    def override(self, **kwargs):
        """
        Returns a context manager that can be used to temporarily set
        attributes in this namespace.

        The context manager has a `delete` method that allows values
        to be removed as well. The method accepts any number of
        attribute names to delete and returns the context manager, so it
        can be used in a :ref:`with` block directly. Missing key names
        are ignored.

        The context manager is a modified version of
        :py:class:`mapping_context`, so it also has a
        :py:meth:`~mapping_context.chain` method.

        All values will be reset to their original values when the
        manager exits. All context managers returned by this method
        operate on the same object, so their effects are cumulative.

        Sample usage::

            n = Namespace(a=1, b=2)
            print(n)
            with n.override(b=0, c=3):
                print(n)
                with n.override(b=1).delete('a'):
                    print(n)
                print(n)
            print(n)

        Will result in::

            Namespace(a=1, b=2)
            Namespace(a=1, b=0, c=3)
            Namespace(b=1, c=3)
            Namespace(b=0, c=3, a=1)
            Namespace(b=2, a=1)
        """
        return _NamespaceOverrideContext(self, **kwargs)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 01 Jun 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.npy_util` module.
"""


import numpy as np
from pytest import raises

from ..npy_util import find_peaks, masked_index, replace_field, unmasked_index


class TestReplaceField:
    inner = np.dtype([('Roll', np.float32),
                      ('Pitch', np.float32),
                      ('Yaw', np.float32)])
    inner2 = np.dtype([('RollPitchYaw', np.float32, 3)])
    outer = np.dtype([('Position', np.float32, 3),
                      ('Attitude', inner)])
    outer2 = np.dtype([('Position', np.float32, 3),
                       ('Attitude', np.float32, 3)])

    def test_replace_custom(self):
        result = replace_field(self.outer, np.float32, 'Attitude')
        expected = self.outer2
        assert result == expected

    def test_replace_with_custom(self):
        result = replace_field(self.outer2, self.inner, 'Attitude')
        assert result == self.outer

    def test_empty_fields(self):
        result = replace_field(self.inner, np.float32)
        expected = self.inner2
        assert result == expected

    def test_name(self):
        dt = np.dtype([('snake', np.float32),
                       ('toSnake', np.float32),
                       ('toCamel', np.float32),
                       ('ToCAMEL', np.float32),
                       ('X', np.float32)])

        result = replace_field(dt, np.float32, name='RPY')
        expected = np.dtype([('RPY', np.float32, 5)])
        assert result == expected

        expected = np.dtype([('snake_toSnake_toCamelToCAMEL_X', np.float32, 5)])
        result = replace_field(dt, np.float32)

    def test_order(self):
        result = replace_field(self.inner, np.float32, 'Roll', 'Pitch', 'Yaw')
        expected = np.dtype([('RollPitchYaw', np.float32, 3)])
        assert result == expected

        result = replace_field(self.inner, np.float32, 'Yaw', 'Pitch', 'Roll')
        expected = np.dtype([('YawPitchRoll', np.float32, 3)])
        assert result == expected

    def test_nested(self):
        result = replace_field(
            self.outer, replace_field(self.inner, np.float32), 'Attitude'
        )
        expected = np.dtype([
            ('Position', np.float32, 3), ('Attitude', self.inner2)
        ])
        assert result == expected

    def test_span(self):
        result = replace_field(self.inner, np.float64, 'Yaw', 'Pitch')
        expected = np.dtype([('Roll', np.float32),
                             ('YawPitch', np.float64)])
        assert result == expected

        raises(ValueError, replace_field, self.inner, np.float64, 'Yaw', 'Roll')

    def test_align(self):
        spec = [('a', np.int32), ('b', np.int8), ('c', np.int32)]
        dt = np.dtype(spec, align=True)

        result = replace_field(dt, np.int8, 'b')
        expected = dt
        assert result == expected

        result = replace_field(dt, np.int8, 'b', 'c')
        expected = np.dtype([('a', np.int32), ('b_c', np.int8, 8)])
        assert result == expected

        dt = np.dtype(spec, align=False)

        result = replace_field(dt, np.int8, 'b')
        expected = dt
        assert result == expected

        result = replace_field(dt, np.int8, 'b', 'c')
        expected = np.dtype([('a', np.int32), ('b_c', np.int8, 5)])
        assert result == expected

    def test_union(self):
        result = replace_field(self.inner, np.float32, 'Yaw', 'Roll')
        expected = np.dtype({
            'names': ['YawRoll', 'Pitch'],
            'formats': [(np.float32, 3), np.float32],
            'offsets': [0, 4],
            'itemsize': 12,
        })
        assert result == expected

    def test_nofields(self):
        raises(ValueError, replace_field, np.dtype(np.float32), np.float32)

    def test_missing(self):
        raises(ValueError, replace_field, self.inner, np.float32, 'XXX')


class TestUnmaskedIndex:
    def test_empty_array(self):
        assert (unmasked_index((), []) == []).all()
        assert (unmasked_index([], []) == []).all()
        raises(IndexError, unmasked_index, 0, [])

    def test_empty_mask(self):
        raises(IndexError, unmasked_index, 0, [False, False, False])

    def test_bad_type(self):
        raises(IndexError, unmasked_index, [1.0, 2.0], np.ones(5, bool))

    def test_non_bool(self):
        expected = [1, 3]
        mask = [1.0, 2.0, 0.0, 3.0]
        result = unmasked_index([1, 2], mask)
        assert (result == expected).all()

    def test_out_of_bounds(self):
        raises(IndexError, unmasked_index, 2, [True, True, False])
        raises(IndexError, unmasked_index, -3, [True, True, False])

    def test_tuple(self):
        expected = [4, 3]
        result = unmasked_index((3, 2), [True, True, False, True, True])
        assert (result == expected).all()

    def test_f_order(self):
        expected = ([0, 1], [1, 2])
        index = [1, 2]
        mask = np.array([[True, True, False], [False, False, True]], order='F')

        result = unmasked_index(index, mask)

        assert len(result) == 2
        for r, e in zip(result, expected):
            assert (r == e).all()

    def test_simple(self):
        expected = [4, 0, 1, 7]
        index = [2, 0, 1, 3]
        mask = [True, True, False, False, True, False, False, True]
        result = unmasked_index(index, mask)
        assert (result == expected).all()

    def test_ndarray(self):
        expected = ([[2, 0], [1, 0]], [[2, 1], [3, 0]])
        index = [[4, 1], [3, 0]]
        mask = [[True, True, False, False],
                [True, False, False, True],
                [False, False, True, False]]
        result = unmasked_index(index, mask)
        assert len(result) == 2
        for r, e in zip(result, expected):
            print(r, e, sep='\n')
            assert (r == e).all()

    def test_negative(self):
        expected = ([[2, 0], [1, 0]], [[2, 1], [3, 0]])
        index = [[-1, -4], [-2, -5]]
        mask = [[True, True, False, False],
                [True, False, False, True],
                [False, False, True, False]]
        result = unmasked_index(index, mask)
        assert len(result) == 2
        for r, e in zip(result, expected):
            assert (r == e).all()


class TestMaskedIndex:
    def test_out_of_bounds(self):
        raises(IndexError, masked_index, 0, [])
        raises(IndexError, masked_index, -3, [True, False])
        raises(IndexError, masked_index, 2, [True, False])
        raises(IndexError, masked_index, (-1, 2), [[True, False]])
        raises(IndexError, masked_index, (1, 0), [[True, False]])

    def test_wrong_dims(self):
        raises(ValueError, masked_index, 0, [[True]])
        raises(ValueError, masked_index, (0, 0, 0), [[True]])

    def test_promotion(self):
        result = masked_index(0, False)
        assert result == -1

        raises(IndexError, masked_index, 1, False)

    def test_not_bool(self):
        result = masked_index(0, [1, 2])
        assert result == 0

    def test_unmasked_portion(self):
        result = masked_index((), False)
        assert result == -1

        result = masked_index(0, [False, True])
        assert result == -1

        result = masked_index((1, 1), [[True, True], [True, False]])
        assert result == -1

    def test_simple(self):
        expected = [0, 2, 3]
        index = [1, 3, 4]
        mask = [False, True, True, True, True, False]
        result = masked_index(index, mask)
        assert (result == expected).all()

    def test_ndarray(self):
        expected = [[0, 8, 1, 2], [9, 11, 3, 7]]
        index = (
            [[0, 1, -2, 0], [-1, 1, -2, -1]],
            [[0, -2, -3, 1], [-2, 2, 1, 0]],
            [[0, -4, 1, -3], [2, -2, 3, -1]]
        )
        mask = [
            [[True, True, False, False],
             [False, True, False, True],
             [False, False, True, True]],
            [[True, False, False, True],
             [True, False, True, False],
             [False, True, True, False]]
        ]
        result = masked_index(index, mask)
        assert (result == expected).all()

    def test_negative(self):
        expected = [0, -1]
        index = ([1, 0], [0, 0])
        mask = [[False], [True]]
        result = masked_index(index, mask)
        assert (result == expected).all()

    def test_broadcast(self):
        expected = [
            [[[0, -1, 1], [0, -1, 1]],
             [[-1, 2, 3], [-1, 2, 3]],
             [[0, -1, 1], [0, -1, 1]]],
            [[[-1, 2, 3], [-1, 2, 3]],
             [[0, -1, 1], [0, -1, 1]],
             [[-1, 2, 3], [-1, 2, 3]]]
        ]
        index = (
            [[[[0]], [[1]], [[0]]], [[[1]], [[0]], [[1]]]],
            [[0, 1, 2], [0, 1, 2]]
        )
        mask = [[True, False, True], [False, True, True]]
        result = masked_index(index, mask)
        assert result.shape == (2, 3, 2, 3)
        assert (result == expected).all(None)


class TestFindPeaks:
    def test_empty(self):
        result = find_peaks([])
        assert (result == []).all()

    def test_single(self):
        result = find_peaks([1.0])
        assert (result == [0]).all()

    def test_flat(self):
        result = find_peaks([0, 1, 1, 1, 1, 1, 0, 2, 2, 2, 2, 3, 3, 3, 3, 3])
        assert (result == [11, 1]).all()

    def test_order_n(self):
        result = find_peaks([0, 2, 1, 3], n_peaks=1)
        assert (result == [3]).all()

        result = find_peaks([0, 2, 1, 3], n_peaks=5)
        assert (result == [3, 1]).all()

    def test_dates(self):
        expected_ind = [3, 5]
        expected_val = np.array(['2022-10-14 10:00:00', '2022-10-14 00:00:00'],
                                dtype='datetime64[ns]')
        dates = np.array([
            '2022-10-13 10:00:00', '2022-10-13 10:30:00',
            '2022-10-13 11:00:00', '2022-10-14 10:00:00',
            '2022-10-13 23:00:00', '2022-10-14 00:00:00'],
            dtype='datetime64[ns]')
        ind, val = find_peaks(dates, mode='value', return_values=True)
        print(ind, expected_ind, sep='\n')
        assert (ind == expected_ind).all()
        assert val.dtype == dates.dtype
        assert (val == expected_val).all()

        ind, val = find_peaks(dates, mode='index', return_values=True)
        assert (ind == expected_ind).all()
        assert val.dtype == dates.dtype
        assert (val == expected_val).all()

    def test_obj(self):
        data = np.empty(5, dtype=object)
        data[0] = []
        data[1] = set()
        data[2] = np.zeros(3)
        data[3] = {}
        data[4] = '1'
        raises(TypeError, find_peaks, data)

    def test_mode(self):
        data = [1, 1, 0, 2, 4, 4, 4, 0, 3, 5]

        raises(ValueError, find_peaks, data, mode='invalid')
        raises(TypeError, find_peaks, data, mode=None)
        raises(TypeError, find_peaks, data, mode=b'value')

        expected = [9, 4, 0]
        result = find_peaks(data, n_peaks=2, mode='value')
        assert (result == expected[:2]).all()
        result = find_peaks(data, mode='vAluE')
        assert (result == expected).all()

        expected = [0, 4, 9]
        result = find_peaks(data, n_peaks=2, mode='index')
        assert (result == expected[:2]).all()
        result = find_peaks(data, mode='INdeX')
        assert (result == expected).all()

    def test_ravel(self):
        data = [[0, 0, 0, 1], [1, 2, 3, 2], [1, 0, 0, 0]]
        expected = [6]
        result = find_peaks(data)
        assert (result == expected).all()

    def test_return_values(self):
        result = find_peaks([0, 3, 0, 1], 2, return_values=True)
        assert len(result) == 2

        ind, val = result
        assert (ind == [1, 3]).all()
        assert (val == [3, 1]).all()

    def test_npeaks(self):
        data = [2, 0, 1, 3, 2, 4, 10, 8, 5, 6, 5, 4, 5]

        expected = [6, 9, 12, 3, 0]
        result = find_peaks(data, mode='value')
        print(result)
        print(expected)
        assert (result == expected).all()
        result = find_peaks(data, 2, mode='value')
        assert (result == expected[:2]).all()

        expected = [0, 3, 6, 9, 12]
        result = find_peaks(data, mode='index')
        assert (result == expected).all()
        result = find_peaks(data, 3, mode='index')
        assert (result == expected[:3]).all()



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 22 Jan 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.numbers` module.
"""


from pytest import raises

from ..numbers import as_base, digit_count


class TestAsBase:
    def test_stack_overflow(self):
        """
        Test from https://stackoverflow.com/a/28666223/2988730, also
        shows that `letters` gets ignored.
        """
        n = 67854 ** 15 - 102
        base = 577
        expected = [4, 473, 131, 96, 431, 285, 524, 486, 28, 23,
                    16, 82, 292, 538, 149, 25, 41, 483, 100, 517,
                    131, 28, 0, 435, 197, 264, 455]
        assert as_base(n, base, letters=True) == expected

    def test_negative(self):
        n = -100
        base = 3
        expected_string = '-10201'
        expected_list = [1, 0, 2, 0, 1]
        assert as_base(n, base, letters=True) == expected_string
        assert as_base(n, base, letters=False) == expected_list

    def test_one(self):
        n = 10
        base = 1
        expected = '1111111111'
        assert as_base(n, base, letters=True) == expected

    def test_non_int(self):
        raises(TypeError, as_base, 0.1)
        raises(ValueError, as_base, 'a')

    def test_negative_base(self):
        raises(ValueError, as_base, 10, -12)


class TestDigitCount:
    def test_one(self):
        n = 100
        assert digit_count(n, 1) == n

    def test_dec(self):
        n = 12877617349487594742394872304230
        assert digit_count(n, 10) == 32

    def test_bin(self):
        n = 0b1010101001000010010011101101
        assert digit_count(n, 2) == 28

    def test_dec(self):
        n = 0o561761237213572357
        assert digit_count(n, 8) == 18

    def test_dec(self):
        n = 12877617349487594742394872304230
        assert digit_count(n, 10) == 32

    def test_hex(self):
        n = 0x12389BE98489F2377BE23ECBD1AEAD3578DB
        assert digit_count(n, 16) == 36

    def test_large(self):
        n = 10**1000
        assert digit_count(n, 10) == 1000

    def test_invalid(self):
        raises(TypeError, digit_count, 8.1)
        raises(TypeError, digit_count, 100, base=7.5)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 22 Jan 2022: Initial Coding

"""
Utilities for manipulating non-computational aspects of numpy arrays.

Mathematical computations belong in :py:mod:`haggis.math`.
"""

__all__ = [
    'find_peaks', 'isolate_dtype', 'iterate_dtype', 'map_array', 'mask2runs',
    'masked_index', 'prune_mask', 'replace_field', 'runs2mask', 'unmasked_index'
]


from itertools import product

import numpy

from . import Sentinel
from .numbers import digit_count


def isolate_dtype(dtype, char='O'):
    """
    Create a new dtype that only contains the fields and subfields
    of `dtype` matching primitive type given by letter `char`.

    The new dtype will be flat with N fields containing the
    offsets to the original elements of interest. It will have the
    same `itemsize` as `dtype`. Field names are not preserved.

    Parameters
    ----------
    dtype : numpy.dtype
        Data type to parse out
    char : str[1]
        Primitive type character code to search for

    Return
    ------
    isolated : dtype
        A dtype containing offsets to all instances of `char` in
        `dtype`.
    """
    def dfs(dtype, base):
        """
        A DFS search will be more likely to yield offsets in the
        correct order without sorting.

        Parameters
        ----------
        dtype :
            The root of the current branch.
        base :
            The offset of dtype within its parent.

        Return
        ------
        offsets :
            The list of offsets of matching primitive fields found
            so far.
        """
        # Structure
        if dtype.fields is not None:
            offsets = []
            for dt, offset in dtype.fields.values():
                offsets.extend(dfs(dt, base + offset))
            return offsets
        # Array
        elif dtype.subdtype is not None:
            offsets = dfs(dtype.subdtype[0], base)
            return list(
                (offsets + dtype.subdtype[0].itemsize *
                 numpy.arange(numpy.prod(dtype.shape))[:, None]).ravel()
            )
        # Primitive
        elif dtype.char == char:
            return [base]
        return []

    offsets = dfs(dtype, 0)
    n = len(offsets)
    k = digit_count(n)
    # Generate names
    names = [f'a{i + 1:0{k}d}' for i in range(n)]
    d = {'names': names, 'offsets': offsets,
         'formats': n * [char], 'itemsize': dtype.itemsize}
    return numpy.dtype(d)


def iterate_dtype(arr, iterate_elements=False, yield_key=False):
    """
    Generate each primitive sub-array of a complex datatype.

    The generator yields the array for each builtin dtype. The leading
    dimensions of each yielded array are `arr.shape` the trailing
    dimensions are determined by `iterate_elements` and the shapes
    present in each sub-dtype. Currently, only depth-first traversal is
    supported.

    An optional field key can be yielded as well, if `yield_key` is
    set. The key is a dot-separated string enumerating the visited
    fields. If `iterate_elements` is specified, it will contain
    bracketed indices as well. See the `Examples` section for more
    information, including a sample of the key format.

    Parameters
    ----------
    arr : numpy.ndarray
        Must have a `dtype` attribute.
    iterate_elements : bool
        If `True`, array elements of each dtype will be yielded
        separately. See `Examples` for more information.
    yield_key : bool
        Whether or not to generate a second output contatining the
        field key.

    Examples
    --------
    Create a complex dtype and an array of zeros::

        >>> dt0 = np.dtype([('a', np.float32), ('b', np.int32, 2)])
        >>> dt = np.dtype([('x', np.bool_), ('y', dt0, 3)])
        >>> arr = np.zeros((3, 3), dt)

    When iterating without elements, the genrator does not descend into
    each sub-dtype consisting of primitives::

        >>> for a, k in iterate_dtype(arr, yield_key=True):
        ...     print(k, a.dtype, a.shape)
        x bool (3, 3)
        y.a float32 (3, 3, 3)
        y.b int32 (3, 3, 3, 2)

    When `iterate_elements` is set, the generator descends into the
    elements of each sub-dtype, even if they are primitive::

        >>> for a, k in iterate_dtype(arr, iterate_elements=True, yield_key=True):
        ...     print(k, a.dtype, a.shape)
        x bool (3, 3)
        y[0].a float32 (3, 3)
        y[0].b[0] int32 (3, 3)
        y[0].b[1]int32 (3, 3)
        y[1].a float32 (3, 3)
        y[1].b[0] int32 (3, 3)
        y[1].b[1] int32 (3, 3)
        y[2].a float32 (3, 3)
        y[2].b[0] int32 (3, 3)
        y[2].b[1] int32 (3, 3)
    """
    def add(key, leaf, index=None):
        key = '{}.{}'.format(key, leaf) if key else leaf
        if index is not None:
            key = '{}[{}]'.format(key, ', '.join(map(str, index)))
        return key

    def pack(arr, key):
        if yield_key:
            return arr, key
        return arr

    if iterate_elements:
        def inner(arr, key):
            dt = arr.dtype
            if dt.fields is None:
                yield pack(arr, key)
            else:
                for fname, (ftype, _) in dt.fields.items():
                    if ftype.subdtype is not None:
                        for index in product(*(range(n) for n in ftype.shape)):
                            yield from inner(arr[fname][(Ellipsis,) + index],
                                             add(key, fname, index))
                    else:
                        yield from inner(arr[fname], add(key, fname))
    else:
        def inner(arr, key):
            dt = arr.dtype
            if dt.fields is None:
                yield pack(arr, key)
            else:
                for field in dt.fields:
                    yield from inner(arr[field], add(key, field))

    yield from inner(arr, '')


def replace_field(in_type, out_type, *fields, name=None):
    """
    Create a dtype that will allow viewing a subset of the fields of
    `in_type` with a different structure.

    This function preserves the names, types, and offsets of all the
    unmodified fields. The replacement type will cover the entirety of
    the named `fields`, regardless of whether the underlying fields are
    contiguous or not.

    If the size of the replaced block is a multiple of
    `out_type.itemsize` other than one, the output type will be an array.
    The multiple must be an integer.

    Parameters
    ----------
    in_type : numpy.dtype
        Datatype to transform.
    out_type : numpy.dtype
        Primitive types may be provided as the equivalent string or
        class object.
    *fields : str
        Names of the fields to transfrom. An empty `fields` means that
        all of them are to be replaced. A new field is generated as a
        contiguous block whose size must be a multiple of
        `out_type.itemsize`. All elements must be valid field names in
        `in_type`.
    name : str, optional
        The name of the replacement field. By default, this is just the
        concatenation of `fields`, respecting CamelCase and snake_case
        conventions in transitions.

    Returns
    -------
    dtype :  numpy.dtype
        Dtype with the named `fields` replaced by a scalar or array of
        `out_type`. All other fields remain the same.

    Examples
    --------
    A simple case:

        >>> inner = np.dtype([('Roll', np.float32),
        ...                   ('Pitch', np.float32),
        ...                   ('Yaw', np.float32)])
        >>> outer = np.dtype([('Position', np.float32, 3),
        ...                   ('Attitude', inner)])
        >>> replace_field(outer, np.float32, 'Attitude')
        dtype([('Position', '<f4', (3,)), ('Attitude', '<f4', (3,))])

    To modify nested custom types, call this function recursively::

        >>> replace_field(outer, replace_field(inner, np.float32), 'Attitude')
        dtype([('Position', '<f4', (3,)), ('Attitude', [('RollPitchYaw', '<f4', (3,))])])
    """
    if in_type.fields is None:
        raise ValueError(f'in_type must be structured dtype, found {in_type}')
    if any(f not in in_type.fields for f in fields):
        raise ValueError('in_type is missing some field(s): '
                         f'{", ".join(set(fields) - set(in_type.fields))}')
    out_type = numpy.dtype(out_type)

    if not fields:
        fields = in_type.fields.keys()
    if not name:
        names = []
        for field in fields:
            if (names and
                ((names[-1][-1].isalpha() ^ field[0].isalpha()) or
                 not (names[-1][-1].isupper() ^ field[0].isupper()))):
                names.append('_')
            names.append(field)
        name = ''.join(names)
    fields = set(fields)

    # Keep track of the elements to retain
    names = []
    formats = []
    offsets = []
    # Keep track of the index of the replacement in `keys`
    idx = None
    # Keep track of the min and max offsets covered
    start = stop = None

    # Generate the new dtype
    for field, (type, offset) in in_type.fields.items():
        if field in fields:
            if idx is None:
                idx = len(names)
                start = offset
                stop = offset + type.itemsize
            else:
                start = min(start, offset)
                stop = max(stop, offset + type.itemsize)
        else:
            names.append(field)
            formats.append(type)
            offsets.append(offset)

    n = (stop - start) // out_type.itemsize
    if out_type.itemsize * n != stop - start:
        raise ValueError(f'Specified block covers {stop - start} bytes, '
                         f'but replacement size is {out_type.itemsize}')

    names.insert(idx, name)
    formats.insert(idx, out_type if n == 1 else (out_type, n))
    offsets.insert(idx, start)
    return numpy.dtype({
        'names': names, 'formats': formats, 'offsets': offsets,
        'itemsize': in_type.itemsize
    })


def map_array(map, arr, value=None, default=Sentinel):
    """
    Convert the elements of a numpy array using a mapping.

    The implementation uses looping to interface between the python and
    numpy datasets, but is as efficient as possible under the
    circumstaces. Intended for mapping a small number of arbitrary
    labels to some alternative value.

    Parameters
    ----------
    map : Mapping
        The mapping to apply. Any object with a `get` method that
        supports default values is accepted.
    arr : array-like
        The array to convert.
    value : callable, optional
        A function to apply to the dictionary values before placing in
        the output array. The default is a no-op.
    default :
        The value to use for array elements not in `mapping`. The
        default is to raise a `KeyError`. `None` is interpreted as a
        valid default.

    Returns
    -------
    mapped_array : array-like
        An array of the same shape as `arr`, with elements transformed
        according to the mapping.
    """
    def get(key):
        v = map.get(key, default)
        if v is Sentinel:
            raise KeyError(key)
        return v

    unq, ind = numpy.unique(arr, return_inverse=True)
    if value:
        vals = [value(get(u)) for u in unq]
    else:
        vals = [get(u) for u in unq]
    return numpy.array(vals)[ind]


def mask2runs(mask, return_lengths=False, return_borders=False):
    """
    Find the runs in a boolean mask.

    Parameters
    ----------
    mask : array-like
        Boolean mask. If not boolean, will be cast to bool.
    return_lengths : bool, optional
        Whether or not to return an array of lengths for each run.
    return_borders : bool, optional
        Whether or not to return an array of dtype `np.int8` containing
        1 at each run start and -1 past run ends. The default is False.

    Returns
    -------
    regions : numpy.ndarray (N, 2)
        Array of indices for each run. First column is the location of
        the run start, second column is past the run end.
    lengths : numpy.ndarray (N)
        The length of each run. This is effectively
        ``regions[:, 1] - regions[:, 0]``. Only returned if
        `return_lengths` is `True`.
    borders : numpy.ndarray (mask.size + 1)
        Array of :py:obj:`numpy.int8` containing 1 at each run start,
        -1 past each run end, and zero elsewhere. Only returned if
        `return_borders` is `True`. Extends one element past the end of
        `mask`. ``np.cumsum(borders).view(bool)[:-1]`` is equivalent to
        `mask`.
    """
    mask = numpy.asanyarray(mask).astype(bool, copy=False)
    borders = numpy.diff(numpy.r_[numpy.int8(0),
                                  mask.view(numpy.int8),
                                  numpy.int8(0)])
    indices = numpy.flatnonzero(borders).reshape(-1, 2)

    if return_lengths:
        lengths = numpy.diff(indices, axis=1).ravel()
        if return_borders:
            return indices, lengths, borders
        return indices, lengths
    elif return_borders:
        return indices, borders
    return indices


def runs2mask(runs, n=None):
    """
    Convert an Nx2 array of run indices, such as the return of
    :py:func:`mask2runs` into a boolean mask of size `n`.

    Parameters
    ----------
    runs : array-like
        A two-column array, the first column being inclusive start
        indices for each run, and the second being exclusive stop
        indices.
    n : int, optional
        The size of the mask to generate. If missing (None), the
        end of the last run is assumed (``runs[-1, 1]``).

    Return
    ------
    mask : numpy.ndarray
        A boolean array of length ``n`` with runs set to True.
    """
    runs = numpy.asanyarray(runs)
    if n is None:
        n = runs[-1, 1]
    mask = numpy.zeros(n, dtype=bool)
    view = mask.view(numpy.int8)

    # Assign start indices
    view[runs[:, 0]] = 1

    # Assign end indices
    ends = runs[:, 1]
    if ends[-1] == n:
        ends = ends[:-1]
    view[ends] = -1

    numpy.cumsum(view, out=view)
    return mask


def prune_mask(mask, min_length=None, max_length=None, filter_func=None,
               return_runs=False, return_lengths=False, return_borders=False):
    """
    Prune the runs in `mask` to conform to some criteria.

    Parameters
    ----------
    mask : array-like
        Boolean mask. If not boolean, will be cast to bool.
    min_length : int, optional
        Runs shorter than this will be removed. Ignored if None.
    max_length : int, optional
        Runs longer than this will be removed. Ignored if None.
    filter_func : callable, optional
        Function that accepts an Nx2 array of runs, as from `mask2runs`,
        and returns an N-length boolean mask. True elements in the
        result will be retained, while False elements will be removed.
        Called after applying `min_length` and `max_length` constraints,
        if those are requested.
    return_runs : bool, optional
        Whether or not to return an Nx2 array of indices for the start
        (inclusive) and end (exclusive), of each run within the mask.
    return_lengths : bool, optional
        Whether or not to return an array of lengths for each retained
        run.
    return_borders : bool, optional
        Whether or not to return an array of dtype `np.int8` containing
        1 at each run start and -1 past run ends.

    Return
    ------
    pruned : numpy.ndarray[bool]
        Mask with any runs shorter than `min_length` and longer than
        `max_length` removed and `filter_func` applied to the remaining runs.
    runs : numpy.ndarray
        Nx2 array of remaining run indices, returned if `return_runs` is set.
        Same format as the result of `mask2runs`. Start index in the first
        column is inclusive, stop index in the second column is exclusive.
    lengths : numpy.ndarray
        N-element array of lengths of remaining runs, returned if
        `return_lengths` is set.
    borders : numpy.ndarray[numpy.int8]
        Array of the same size as `mask` containing 1 at the start of each run
        and -1 past the end, returned if `return_borders` is set. The sum of
        this array is the mask.
    """
    def len_mask(mask):
        nonlocal runs, lengths
        borders[runs[~mask, :]] = 0
        runs = runs[mask]
        lengths = lengths[mask]

    runs, lengths, borders = mask2runs(mask,
                                       return_lengths=True, return_borders=True)
    if min_length is not None:
        len_mask(lengths >= min_length)
    if max_length is not None:
        len_mask(lengths <= max_length)
    if filter_func is not None:
        len_mask(filter_func(runs))

    mask = numpy.cumsum(borders[:-1], out=borders[:-1]).view(bool)
    result = [mask]
    if return_runs:
        result.append(runs)
    if return_lengths:
        result.append(lengths)
    if return_borders:
        result.append(borders)
    if len(result) == 1:
        return mask
    return tuple(result)


def unmasked_index(index, mask):
    """
    Convert `index` in a masked aray into the corresponding index in the
    original.

    Given an array ``x`` of the same size as `mask`, determine the
    position of ``x[mask][index]`` in the original array ``x``. Since
    `mask` may be multi-dimensional but ``x[mask]`` is always raveled,
    the result may be a scalar or a tuple.

    Parameters
    ----------
    index : array-like[int]
        The index or indices into in the masked (raveled) array. Each
        element must be in
        ``[-np.count_nonzero(mask), np.count_nonzero(mask))``.
    mask : array-like[bool]
        A boolean mask to determine the indices of the array.
        Non-boolean mask arrays will be interpreted as booleans, as
        though with ``mask.astype(bool)``. Scalar arrays will be
        interpreted as one-element arrays, so scalar indexing 
        allowed.

    Return
    ------
    unmasked_index : int or numpy.ndarray or tuple
        If `mask` is a scalar or 1D, the result is the same size and
        shape as `index`. Otherwise, it is a tuple of length
        ``mask.ndim`` each of whose elements corresponds to a dimension,
        and it the same size and shape as `index``.
    """
    index = numpy.asanyarray(index)
    mask = numpy.asanyarray(mask)

    if index.size == 0:
        empty = numpy.array([], dtype=numpy.intp)
        return (empty,) * mask.ndim if mask.ndim > 1 else empty
    indices = numpy.argwhere(mask)
    if mask.ndim > 1:
        return tuple(indices.T[:, index])
    indices = indices[index].ravel()
    return indices if index.ndim else indices.item()


def masked_index(index, mask):
    """
    Convert N-dimensional `index` into its corresponding location in the
    masked array, if it is in the masked portion.

    Given an array ``x``, of the same size as `mask`, determine the
    position of ``x[index]`` in ``x[mask]``. If `index` is not in the
    masked portion, the corressponding output is set to ``-1``.

    Parameters
    ----------
    index : tuple
        The index or indices into the unmasked multi-dimensional array.
        A non-tuple (scalar or array) index may be used for 1D masks,
        will be wrapped in a tuple internally to avoid ambiguity. The
        number of elements in the tuple must match ``mask.ndim``, and
        all elements must all broadcast together.
    mask : array-like[bool]
        A boolean mask to determine the indices of the array.
        Non-boolean mask arrays will be interpreted as booleans, as
        though with ``mask.astype(bool)``. Scalar arrays will be
        interpreted as one-element arrays, so scalar indexing 
        allowed.

    Return
    ------
    masked_index : int or None
        If `index` corresponds to a `True` element of `mask`, return its
        position in the masked result. Otherwise, return `None`.
    """
    mask = numpy.asanyarray(mask, dtype=numpy.bool_)
    if not isinstance(index, tuple):
        index = (index,)
    if mask.ndim > 0 and mask.size > 0:
        if mask.ndim != len(index):
            raise ValueError('index must have length {}'.format(mask.ndim))
        # The following two lines are a hack until a new mode is added. See
        # https://github.com/numpy/numpy/issues/15475#issuecomment-1280056626
        index = tuple(numpy.asanyarray(i) for i in index)
        index = tuple(numpy.where(i < 0, i + s, i)
                            for i, s in zip(index, mask.shape))
        try:
            index = numpy.ravel_multi_index(index, mask.shape, mode='raise')
        except ValueError as e:
            raise IndexError(str(e))
    mask = mask.ravel()
    indices = mask.cumsum(dtype=numpy.intp)
    indices -= 1
    return numpy.where(mask[index], indices[index], numpy.intp(-1))


def find_peaks(arr, n_peaks=None, mode='value', return_values=False):
    """
    Find the locations of the `n_peaks` tallest or leftmost (unfiltered)
    local maxima of `arr`.

    When ``mode=='value'``, the first peak is located at
    ``arr.argmax()``. The second peak is the maximum value among the
    numbers that are not monotonically non-increasing away from the
    first peak. Successive peaks are returned from remaining portions
    of the array.

    For a plateau with multiple equal peak elements, only the first is
    returned. Similarly, if multiple peaks with the same values are
    found, they will be retreived in order of increasing index.

    Parameters
    ----------
    arr : array_like
        The data to search through. Expected to be 1D. Larger dimensions
        are raveled to avoid the problem of different numbers of results
        along a given axis.
    n_peaks : int or None
        The number of peaks to search for. None or negative searches for
        all available peaks. If `n_peaks` local maxima are not found,
        the result will be shorter than requested.
    mode : str
        One of {``'value'``, ``'index'``} (case insensitive):

        - ``'value'``: Return up to the first `n_peaks` tallest peaks,
          regardless of location.
        - ``'index'``: Return up to the first `n_peaks` local maxima,
          traversing the array from left to right, regardless of peak
          height.
    return_values : bool
        Return a second array containing the maximum values at each
        location. Convenience for ``arr[find_peaks(arr, ...)]``.

    Return
    ------
    peak_indices : numpy.ndarray[int]
        Locations of up to `n_peaks` local maxima. May be shorter than
        `n_peaks` if there are insufficient local maxima in the data.
    peak_values : numpy.ndarray
        Elements of `arr` at `peak_indices`.

    Notes
    -----
    To get peaks in reverse order with ``mode='index'``, reverse the
    input: ``arr[::-1]``.

    To get peaks sorted from left to right with ``mode='value'``, sort
    the results.
    """
    arr = numpy.asanyarray(arr).ravel()
    n = len(arr)
    if n_peaks is None:
        n_peaks = -1

    if not isinstance(mode, str):
        raise TypeError('mode must be one of the strings {{valid, index}}, '
                        'not {}'.format(type(mode).__name__))
    cmode = mode.casefold()

    indices = []
    if cmode == 'value':
        mask = numpy.ones(arr.shape, dtype=numpy.bool_)
        while n and len(indices) != n_peaks:
            index = unmasked_index(arr[mask].argmax(), mask)
            indices.append(index)
            # Find surrounding window
            start = mask[index::-1].argmin() # First zero before
            if start != 0:
                start = index - (start - 1)
            end = mask[index:].argmin()      # First zero after
            if end == 0:
                end = len(arr)
            else:
                end += index
            index -= start
            diff = numpy.diff(arr[start:end])
            zero = diff.dtype.type(0)
            # Find left side start of peak
            if index > 1:
                sstart = (diff[index - 1::-1] < zero).argmax()
                # If index is zero, entire left side is non-increasing
                if sstart != 0:
                    # Otherwise, adjust for reversal
                    sstart = index - sstart
            else:
                sstart = 0
            # Find right side end of peak
            if index < end - start - 2:
                send = (diff[index:] > zero).argmax()
                # If index is zero, entire right side is non-increasing
                if send != 0:
                    end = start + index + send + 1
            start += sstart
            mask[start:end] = False
            n -= end - start
    elif cmode == 'index':
        subset = arr
        offset = 0
        while n and len(indices) != n_peaks:
            if n == 1:
                index = 0
                end = 1
            else:
                diff = numpy.diff(subset)
                zero = diff.dtype.type(0)
                end = (diff < zero).argmax()
                if diff[end] >= 0:
                    # end == 0, since subset is constantly rising, set to last
                    end = n
                    index = n - 1
                else:
                    # Find beginning of plateau
                    index = (diff[end::-1] > zero).argmax()
                    # If index is zero, entire subset is plateau, leave it at 0
                    if index != 0:
                        # Otherwise, adjust for reversal
                        index = end - (index - 1)
            # Found a peak
            indices.append(index + offset)
            # Find end of trough
            if end < n - 2:
                end += (diff[end + 1:] > zero).argmax() + 1
                if diff[end] <= 0:
                    # Trough never ends
                    end = n
                else:
                    # Select first element of uptrend
                    end += 1
            else:
                end = n
            # Update subset
            offset += end
            n -= end
            subset = subset[end:]
    else:
        raise ValueError('mode must be one of {{valid, index}}, '
                         'not {}'.format(mode))

    indices = numpy.array(indices, dtype=numpy.intp)
    if return_values:
        return indices, arr[indices]
    return indices



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for working with new-style MS Word documents, only available
when the ``[docx]`` :ref:`extra <installation-extras>` is installed.

If `python-docx`_ is not found at import time, this module will have a
:py:data:`docx_enabled` attribute, which will be :py:obj:`False`. If
`python-docx`_ is found, on the other hand, :py:data:`docx_enabled` will
be :py:obj:`True`, and all the functions and attributes of the module
will be present.

Some of the methods here are workarounds for features that are missing
or buggy in the original library. Often, these methods are inspired by
recipes found in the corresponding bug reports and Stack Overflow posts,
which are referenced as appropriate.

.. py:data:: docx_enabled

   A boolean value indicating whether the ``[docx]``
   :ref:`extra <installation-extras>` has been installed. If
   :py:obj:`False`, the API will be severely limited.


.. include:: /link-defs.rst 
"""


__all__ = ['docx_enabled',]


from collections import OrderedDict
from os.path import dirname, join
import re

try:
    from docx.document import Document
    from docx.enum.style import WD_STYLE_TYPE
    from docx.enum.section import WD_SECTION_START, WD_ORIENTATION
    from docx.oxml.shared import OxmlElement, qn
    from docx.oxml.table import CT_Tbl
    from docx.oxml.text.paragraph import CT_P
    from docx.shared import RGBColor, Inches, Cm, Mm, Pt, Twips, Emu
    from docx.table import _Cell, Table
    from docx.text.run import Run
    from docx.text.paragraph import Paragraph

    from lxml import etree
except ImportError:
    from .. import _display_missing_extra
    _display_missing_extra('docx', ['python-docx', 'lxml'])
    docx_enabled = False
else:
    __all__.extend([
        'block_iterator', 'insert_toc', 'merge_row', 'set_row_height',
        'style_row', 'style_column', 'table_no_fill', 'edit_font',
        'is_paragraph_empty', 'delete_paragraph', 'list_number', 'add_section',
        'insert_math_ml', 'str2length',
    ])
    docx_enabled = True


if docx_enabled:
    def block_iterator(parent):
        """
        Yield each paragraph and table child within `parent`, in
        document order.

        Each returned value is an instance of either
        :py:class:`~docx.table.Table` or
        :py:class:`~docx.text.paragraph.Paragraph`. `parent` would most
        commonly be a reference to a main
        :py:class:`~docx.document.Document` object, but also works for a
        :py:class:`~docx.table._Cell` object, which itself can contain
        paragraphs and tables.

        This function is taken verbatim from python-docx
        `Issue #40 <https://github.com/python-openxml/python-docx/issues/40>`_.
        Hopefully it will make it to `python-docx`_ soon.


        .. include:: /link-defs.rst 
        """
        if isinstance(parent, Document):
            parent_elm = parent.element.body
        elif isinstance(parent, _Cell):
            parent_elm = parent._tc
        else:
            raise ValueError("something's not right")

        for child in parent_elm.iterchildren():
            if isinstance(child, CT_P):
                yield Paragraph(child, parent)
            elif isinstance(child, CT_Tbl):
                yield Table(child, parent)


    def insert_toc(doc, min_level=1, max_level=3):
        """
        Insert a table of contents stub into a
        :py:class:`~docx.document.Document`.

        The TOC is a stub and needs to be updated manually when the
        end-user opens the generated document in a Word client.

        This function is taken almost verbatim from @mustash's comment
        to python-docx `Issue #36 <https://github.com/python-openxml/python-docx/issues/36#issuecomment-145302669>`_.
        See the previous comment in the thread for more information.
        """
        paragraph = doc.add_paragraph()
        run = paragraph.add_run()
        fld_char = OxmlElement('w:fldChar')          # creates a new element
        fld_char.set(qn('w:fldCharType'), 'begin')   # sets attribute on element
        instr_text = OxmlElement('w:instrText')
        instr_text.set(qn('xml:space'), 'preserve')  # sets attribute on element
        instr_text.text = 'TOC \\o "{min}-{max}" \\h \\z \\u'.format(
                            min=int(min_level), max=int(max_level))

        fld_char2 = OxmlElement('w:fldChar')
        fld_char2.set(qn('w:fldCharType'), 'separate')
        fld_char3 = OxmlElement('w:t')
        fld_char3.text = 'Right-click to update field.'
        fld_char2.append(fld_char3)

        fld_char4 = OxmlElement('w:fldChar')
        fld_char4.set(qn('w:fldCharType'), 'end')

        r_element = run._r
        r_element.append(fld_char)
        r_element.append(instr_text)
        r_element.append(fld_char2)
        r_element.append(fld_char4)


    def merge_row(table, row=0):
        """
        Merge a row in a :py:class:`~docx.table.Table` into a single
        cell.

        Operates on the first row by default.
        """
        rc = table.row_cells(row)
        rc[0].merge(rc[-1])


    def set_row_height(row, height):
        """
        Set the height of a :py:class:`docx.table.Table` row.

        Parameters
        ----------
        row : docx.table._Row
            The row index.
        height : docx.shared.Length
            The height to assign.


        This function will be obsolted by python-docx
        `Pull Request #301 <https://github.com/python-openxml/python-docx/pull/301>`_.
        """
        trHeight = OxmlElement('w:trHeight')
        trHeight.set(qn('w:val'), str(height.twips))

        trPr = OxmlElement('w:trPr')
        trPr.append(trHeight)

        row._tr.append(trPr)


    def style_row(doc, style, *args):
        """
        .. py:function:: style_row(doc, style, row)
           :noindex:
        .. py:function:: style_row(doc, style, table, rowid)
           :noindex:

        Set a custom style for all the text in a row in the table.

        Every paragraph of each cell in the row will be styled, so use
        with care.

        Parameters
        ----------
        doc : docx.document.Document
            The document containing the items to style. Only required 
            when `style` is set by name.
        style : docx.styles.style._CharacterStyle or docx.styles.style._ParagraphStyle or str
            The name or style object representing a run- or
            paragraph-level style to apply.
        row : docx.table._Row
            The row object to style.
        table : docx.table.Table
            The table to style.
        rowid : int
            The index of the row to style.
        """
        if len(args) == 1:
            row, = args
        elif len(args) == 2:
            table, rowid = args
            row = table.rows[rowid]
        else:
            raise TypeError(
                'Expected 1 or 2 positional arguments, got {}'.format(len(args))
            )

        if isinstance(style, str):
            style = doc.styles[style]

        if style.type == WD_STYLE_TYPE.PARAGRAPH:
            def apply(par, style):
                par.style = style
        elif style.type == WD_STYLE_TYPE.CHARACTER:
            def apply(par, style):
                for run in par.runs:
                    run.style = style
        else:
            raise ValueError('Paragraph and character styles only.')

        for cell in row.cells:
            for par in cell.paragraphs:
                apply(par, style)


    def style_column(doc, style, *args):
        """
        .. py:function:: style_column(col, *, style)
           :noindex:
        .. py:function:: style_column(table, colid, *, style)
           :noindex:

        Set a custom style for all the text in a column in the table.

        Every paragraph of each cell in the column will be styled, so
        use with care.

        Parameters
        ----------
        doc : docx.document.Document
            The document containing the items to style. Only required
            when `style` is set by name.
        style : docx.styles.style._CharacterStyle or docx.styles.style._ParagraphStyle or str
            The name or style object representing a run- or
            paragraph-level style to apply.
        col : docx.table._Column
            The column object to style.
        table : docx.table.Table
            The table to style.
        colid : int
            The index of the column to style.
        """
        if len(args) == 1:
            col, = args
        elif len(args) == 2:
            table, colid = args
            col = table.columns[colid]
        else:
            raise TypeError(
                'Expected 1 or 2 positional arguments, got {}'.format(len(args))
            )

        if isinstance(style, str):
            style = doc.styles[style]

        if style.type == WD_STYLE_TYPE.PARAGRAPH:
            def apply(par, style):
                par.style = style
        elif style.type == WD_STYLE_TYPE.CHARACTER:
            def apply(par, style):
                for run in par.runs:
                    run.style = style
        else:
            raise ValueError('Paragraph and character styles only.')

        for cell in col.cells:
            for par in cell.paragraphs:
                apply(par, style)


    def table_no_fill(table):
        """
        Set the table not to fill up the entire page width.

        This should be called after the table has been filled in, or at
        least the number of rows and columns has been set. Adding rows
        and columns may cause unpredictable layout conflicts which may
        invalidate these settings.

        This method is a result of research done on
        https://github.com/python-openxml/python-docx/issues/315.
        """
        for row in table.rows:
            for col in row._tr.tc_lst:
                tcW = col.tcPr.tcW
                tcW.type = 'auto'
                tcW.w = 0


    #: Mapping of supported :py:class:`~docx.text.run.Font` attributes
    #: to dictionaries describing how to set them. All nested keys are
    #: optional. Options are:
    #:
    #:   - ``parser``: Method that converts inputs to values
    #:     `python-docx`_ understands.
    #:   - ``formatter``: Method that converts inputs from internal
    #:     `python-docx`_ values to user-friendly ones.
    #:   - ``setter``: Alternative method to set the attribute on the
    #:     :py:class:`~docx.text.run.Font` object.
    #:
    #: Used by :func:`edit_font`.
    _known_font_attributes = OrderedDict([
        ('name', {'formatter': lambda x: ' + "{}"'.format(x)}),
        ('size', {
            'parser': lambda x: Pt(x),
            'formatter': lambda x: ' + {}pt'.format(x.pt),
        }),
        ('color', {
            'parser': lambda x: RGBColor(*x),
            'formatter': lambda x: ' + rgb({}, {}, {})'.format(*x),
            'setter': lambda font, attr, x: setattr(font.color, 'rgb', x)
        }),
        ('all_caps', {}),
        ('bold', {}),
        ('double_strike', {}),
        ('emboss', {}),
        ('imprint', {}),
        ('italic', {}),
        ('outline', {}),
        ('shadow', {}),
        ('small_caps', {}),
        ('strike', {}),
        ('subscript', {}),
        ('superscript', {}),
        ('underline', {}),
    ])


    def edit_font(doc, runs, **kwargs):
        """
        Adds a dynamic run-style to the document that changes the font
        properties of the specified runs.

        The updated style is generated on the fly and inherits all
        properties except the ones that are modified from the current
        character style of the run. The style name is the original style
        name with the modified properties appended after a ``+`` or
        ``-`` sign, in alphabetical order. If such a character style
        already exists in the document, it will be used instead of
        making a new one.

        The `runs` parameter is expected to generally come from the
        :py:attr:`~docx.text.paragraph.Paragraph.runs` attribute of a
        :py:class:`~docx.text.paragraph.Paragraph` object. A single
        :py:class:`~docx.text.run.Run` can be processed as well.

        Parameters
        ----------
        doc : docx.document.Document
            The document containing the runs to modify. This is
            necessary because the styles must be added to the document.
        runs : iterable[docx.text.run.Run] or docx.text.run.Run
            Either an iterable or runs or a single run. The runs will be
            given a new character style that inherits from the current
            one but contains an updated font with the requested color.
        name: str
            The name of the font. :py:obj:`None` if not to be modified.
            Added as ``... + "NAME"``.
        size: float or int
            The font size in points. :py:obj:`None` if not to be
            modified. Added as ``... + SIZEpt``.
        color: tuple(int, int, int)
            A three-element tuple of integers that represents the RGB
            components of the color to set, or `None` if color is not to
            be modified. Other iterables are accepted, but they are
            converted to tuples internally. Added as
            ``... + rgb(R,G,B)``.


        All remaining arguments are processed as tri-state booleans,
        where :py:obj:`None` indicates no modification. Flags are
        appended as ``... + FLAG`` if :py:obj:`True`, ``... - FLAG`` if
        :py:obj:`False`. Supported arguments flags are: 

          - `bold`
          - `italic`
          - `underline`
          - `subscript`
          - `superscript`
          - `all_caps`
          - `emboss`
          - `strike`
          - `double_strike`
          - `imprint`
          - `outline`
          - `shadow`
          - `small_caps`
        """
        if runs is None:
            return

        if isinstance(runs, Run):
            return edit_font(doc, [runs], **kwargs)

        for run in runs:
            old_style = run.style
            if old_style is None:
                old_style = doc.styles.default(WD_STYLE_TYPE.CHARACTER)
            new_name = old_style.name
            for attr in kwargs:
                value = kwargs[attr]
                if value is not None:
                    properties = _known_font_attributes[attr]
                    if 'parser' in properties:
                        value = properties['parser'](value)
                        kwargs[attr] = value
                    if 'formatter' in properties:
                        new_name += properties['formatter'](value)
                    else:
                        new_name += ' {} {}'.format('+' if value else '-', attr)
            try:
                new_style = doc.styles[new_name]
            except KeyError:
                new_style = doc.styles.add_style(new_name, WD_STYLE_TYPE.CHARACTER)
            new_style.base_style = old_style
            font = new_style.font
            for attr, value in kwargs.items():
                properties = _known_font_attributes[attr]
                if 'setter' in properties:
                    properties['setter'](font, attr, value)
                else:
                    setattr(font, attr, value)
            run.style = new_style


    def is_paragraph_empty(paragraph):
        """
        Check if a paragraph contains content.

        Content includes any item besides styling information and empty runs.

        Source: https://stackoverflow.com/q/51217113/2988730
        """
        p = paragraph._p
        runs = p.xpath('./w:r[./*[not(self::w:rPr)]]')
        others = p.xpath('./*[not(self::w:pPr) and not(self::w:r) and '
                         'not(contains(local-name(), "bookmark"))]')
        return len(runs) + len(others) == 0


    def delete_paragraph(paragraph):
        """
        Remove a paragraph from an enclosing document.

        This may or may not work properly if the paragraph contains
        non-trivial content, like pictures.

        Inspiration is from @scanny's comment regarging python-docx
        `Issue #33 <https://github.com/python-openxml/python-docx/issues/33#issuecomment-77661907>`_.
        """
        p = paragraph._element
        p.getparent().remove(p)
        p._p = p._element = None


    def list_number(doc, par, prev=None, level=None, num=True):
        """
        Make a paragraph into a list item with a specific level and
        optional restart.

        An attempt will be made to retreive an abstract numbering style
        that corresponds to the style of the paragraph. If that is not
        possible, the default numbering or bullet style will be used
        based on the `num` parameter.

        Parameters
        ----------
        doc : ~docx.document.Document
            The document to add the list into.
        par : ~docx.text.paragraph.Paragraph
            The paragraph to turn into a list item.
        prev : ~docx.text.paragraph.Paragraph or None
            The previous paragraph in the list. If specified, the
            numbering and styles will be taken as a continuation of this
            paragraph. If omitted, a new numbering scheme will be
            started.
        level : int or None
            The level of the paragraph within the outline. If `prev` is
            set, defaults to the same level as in `prev`. Otherwise,
            defaults to zero.
        num : bool
            If `prev` is :py:obj:`None` and the style of the paragraph
            does not correspond to an existing numbering style, this
            will determine wether or not the list will be numbered or
            bulleted. The result is not guaranteed, but is fairly safe
            for most Word templates.

        Note
        ----
        The code here is mainly taken from python-docx
        `Issue #25 <https://github.com/python-openxml/python-docx/issues/25>`_
        and `Pull Request #110 <https://github.com/python-openxml/python-docx/pull/110>`_
        In particular, see the two comments by @yurac:
        `[1] <https://github.com/python-openxml/python-docx/issues/25#issuecomment-140334543>`_
        and `[2] <https://github.com/python-openxml/python-docx/issues/25#issuecomment-143231954>`_
        """
        xpath_options = {
            True: {'single': 'count(w:lvl)=1 and ', 'level': 0},
            False: {'single': '', 'level': level},
        }

        def style_xpath(prefer_single=True):
            """
            The style comes from the outer-scope variable ``par.style.name``.
            """
            style = par.style.style_id
            return (
                'w:abstractNum['
                    '{single}w:lvl[@w:ilvl="{level}"]/w:pStyle[@w:val="{style}"]'
                ']/@w:abstractNumId'
            ).format(style=style, **xpath_options[prefer_single])

        def type_xpath(prefer_single=True):
            """
            The type is from the outer-scope variable ``num``.
            """
            type = 'decimal' if num else 'bullet'
            return (
                'w:abstractNum['
                    '{single}w:lvl[@w:ilvl="{level}"]/w:numFmt[@w:val="{type}"]'
                ']/@w:abstractNumId'
            ).format(type=type, **xpath_options[prefer_single])

        def get_abstract_id():
            """
            Select as follows:

                1. Match single-level by style (get min ID)
                2. Match exact style and level (get min ID)
                3. Match single-level decimal/bullet types (get min ID)
                4. Match decimal/bullet in requested level (get min ID)
                3. 0
            """
            for fn in (style_xpath, type_xpath):
                for prefer_single in (True, False):
                    xpath = fn(prefer_single)
                    ids = numbering.xpath(xpath)
                    if ids:
                        return min(int(x) for x in ids)
            return 0

        if (prev is None or
                prev._p.pPr is None or
                prev._p.pPr.numPr is None or
                prev._p.pPr.numPr.numId is None):
            if level is None:
                level = 0
            numbering = doc.part.numbering_part.numbering_definitions._numbering
            # Compute the abstract ID first by style, then by num
            anum = get_abstract_id()
            # Set the concrete numbering based on the abstract numbering ID
            numbr = numbering.add_num(anum)
            # Make sure to override the abstract continuation property
            numbr.add_lvlOverride(ilvl=level).add_startOverride(1)
            # Extract the newly-allocated concrete numbering ID
            numbr = numbr.numId
            ret = numbr, anum
        else:
            # Get the previous concrete numbering ID
            if level is None:
                level = prev._p.pPr.numPr.ilvl.val
            numbr = prev._p.pPr.numPr.numId.val
            ret = numbr

        if num:
            par._p.get_or_add_pPr().get_or_add_numPr().get_or_add_numId().val = numbr
        par._p.get_or_add_pPr().get_or_add_numPr().get_or_add_ilvl().val = level
        return ret


    def add_section(doc, orientation=WD_ORIENTATION.PORTRAIT):
        """
        Add a new section to `doc` with the specified page orientation.

        This function always creates a new section with page break style
        :ref:`wdsectionstart`.NEW_PAGE, even if
        the previous section has the same orientation. The width and
        height of the new section will be swapped if necessary so that
        the width is greater in landscape mode and the height is greater
        in portrait mode.

        `orientation` may be one of the :ref:`wdorientation` enums, or
        the strings {``'portrait'``, ``'landscape'``} (case
        insensitive).

        Returns the newly created section.

        This function is a workaround for python-docx bug
        `#214 <https://github.com/python-openxml/python-docx/issues/214>`_.
        The same workaround is also described in
        http://stackoverflow.com/q/31893557/2988730.
        """
        if isinstance(orientation, str):
            orientation = orientation.lower()
            if orientation == 'portrait':
                orientation = WD_ORIENTATION.PORTRAIT
            elif orientation == 'landscape':
                orientation = WD_ORIENTATION.LANDSCAPE
            else:
                raise ValueError(
                        'Invalid page orientation "{}"'.format(orientation)
                )
        section = doc.add_section(WD_SECTION_START.NEW_PAGE)
        section.orientation = orientation
        width, height = section.page_width, section.page_height
        if (orientation == WD_ORIENTATION.PORTRAIT and width > height) or \
           (orientation == WD_ORIENTATION.LANDSCAPE and height > width):
            section.page_width, section.page_height = height, width
        return section


    try:
        # realpath/abspath omitted intentionally
        _mml2omml_file = join(dirname(__file__), '_resources', 'MML2OMML.XSL')
        _mml2omml_xslt = etree.parse(_mml2omml_file)
    except (OSError, FileNotFoundError):
        def insert_math_ml(*args, **kwargs):
            raise NotImplementedError(
                'Unable to find XSL file at "{}"'.format(_mml2omml_file)
            )
    else:
        _mml2omml_transform = etree.XSLT(_mml2omml_xslt)
        def insert_math_ml(par, math_ml):
            """
            Convert a MathML equation to an Open MathML format suitable
            for MS documents, and insert it into the specified
            paragraph.

            The MathML is converted to Open MathML format using an
            internal stylesheet and inserted into the end of the
            paragraph.

            Parameters
            ----------
            par : ~docx.text.paragraph.Paragraph
                The paragraph to append the equation to.
            math_ml : str or file-like
                If a string that starts with an opening ``<math>`` tag
                and ends with a closing ``</math>`` tag, it will be
                parsed as MathML. Other wise it will be interpreted as a
                file name or file object, as appropriate.

            Notes
            -----
            If the MML to OMML stylesheet can not be located for any
            reason, this function will raise a
            :py:exc:`NotImplementedError`. The error will indicate the
            expected location of the file.

            This function and its setup are based on the discussion of
            `issue #320 <https://github.com/python-openxml/python-docx/issues/320>`_
            on the python-docx GitHub page. The file MML2OMML.XSL was provided
            by user @peepall as part of the discussion.
            """
            loader = etree.parse
            if isinstance(math_ml, str):
                mmc = math_ml.strip().casefold()
                if mmc.startswith('<math') and mmc.endswith('</math>'):
                    loader = etree.fromstring

            mml = loader(math_ml)
            omml = _mml2omml_transform(mml)
            par._element.append(omml.getroot())


    #: Mapping of recognized unit names to their corresponding
    #: `python-docx`_ classes.
    _units_mapping = {
        'emu': Emu,
        'english metric unit': Emu,
        'english metric units': Emu,
        'englishmetricunit': Emu,
        'englishmetricunits': Emu,
        'cm': Cm,
        'centimeter': Cm,
        'centimeters': Cm,
        '"': Inches,
        'in': Inches,
        'inch': Inches,
        'inches': Inches,
        'millimeter': Mm,
        'millimeters': Mm,
        'mm': Mm,
        'pt': Pt,
        'point': Pt,
        'points': Pt,
        'twip': Twips,
        'twips': Twips,
    }


    #: Regular expression to extract floats from a string.
    _float_pattern = re.compile(
        r'[+-]?(?:(?:\d*\.\d+)|(?:\d+\.?))(?:[Ee][+-]?\d+)?'
    )


    def str2length(string, default_units='"'):
        """
        Convert a string with a floating point number and associated
        units into a :py:class:`docx.shared.Length`.

        The number and the units may be separated by optional
        whitespace. If the units are missing entirely, `default_units`
        will be used instead. `default_units` may be a
        :py:class:`docx.shared.Length` subclass, or a string naming the
        units. The default default is :py:class:`docx.shared.Inches`.

        Unit names are case insensitive. They may be written out fully or
        abbreviated, plural or singular.
        """
        string = str(string)
        match = _float_pattern.match(string)
        if match is None:
            raise ValueError('Must begin with valid floating point component')
        num = float(match.group())
        units = string[match.end():].strip()

        if not units:
            units = default_units

        if isinstance(units, str):
            cunits = units.casefold()
            if cunits not in _units_mapping:
                raise ValueError('Unknown units {!r}'.format(units))
            units = _units_mapping[cunits]

        return units(num)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Conversion utilities for PDF files to other formats.

This module relies on the ``[pdf]`` :ref:`extra <installation-extras>`,
which implies external programs. As such, this module may be quite
OS-sensitive. Specifically, it requires the programs :program:`pdftoppm`
and `ImageMagick`_\ 's :program:`convert`.

A small import-guarded block is provided to demo
:py:func:`pdf_to_image`.


.. include:: /link-defs.rst
"""

__all__ = ['pdftoppm_exe', 'convert_exe', 'pdf_to_image']


import subprocess, io


#: The name of the :program:`pdftoppm` executable. Either a full path,
#: or a program that the shell can find on the :envvar:`PATH` is
#: necessary.
pdftoppm_exe = 'pdftoppm'


#: The name of the `ImageMagick`_ :program:`convert` executable. Either
#: a full path, or a program that the shell can find on the
#: :envvar:`PATH` is necessary.
convert_exe = 'convert'


def pdf_to_image(input_path, output_path, format=None):
    """
    Convert a PDF document into an image file.

    This function uses the :py:mod:`subprocess` module to operate. It
    requires the presence of the :program:`pdftoppm` program as well as
    :program:`convert` from `ImageMagick`_.

    `input_path` may be a string path or a file-like object.

    `output_path` may be a string, a file-like object or :py:obj:`None`.
    If :py:obj:`None`, an :py:class:`io.BytesIO` object is returned
    containing the image. `format` defaults to ``'png'`` if not set
    explicitly in this case.

    Return the name of the output file, or an in-memory file-like object
    (:py:class:`io.BytesIO`) if `output_path` is :py:obj:`None`.

    The idea for behind this conversion mechanism comes from 
    http://stackoverflow.com/a/2002436/2988730. The implementation
    details are described in
    http://stackoverflow.com/a/4846923/2988730.


    .. include:: /link-defs.rst
    """
    if isinstance(input_path, str):
        step1 = [pdftoppm_exe, input_path]
        stdin = None
    else:
        step1 = [pdftoppm_exe]
        stdin = input_path

    if output_path is None:
        # Output to memory
        out = '-'
        if format is None:
            format = 'png'
    elif isinstance(output_path, str):
        # Output to path
        out = output_path
    else:
        # Output to file-like object
        out = '-'

    if format is not None:
        out = ':'.join((format, out))

    step2 = [convert_exe, '-', out]

    proc1 = subprocess.Popen(step1, stdin=stdin, stdout=subprocess.PIPE)
    proc2 = subprocess.Popen(step2, stdin=proc1.stdout, stdout=subprocess.PIPE)
    # Force error in case proc2 dies unexpectedly
    proc1.stdout.close()
    # Ignore output
    sysout, _ = proc2.communicate()

    if isinstance(output_path, str):
        return output_path
    elif output_path is None:
        outputFile = io.BytesIO()
        rewind = True
    else:
        outputFile = output_path
        rewind = False
    outputFile.write(sysout)
    if rewind:
        outputFile.seek(0)
    return outputFile

if __name__ == '__main__':
    from sys import argv, stderr, exit
    from shutil import copyfileobj

    if len(argv) <= 1:
        print('Usage: python -m haggis.files.pdf input.pdf '
              '[output.* | ""] [format]', file=stderr)
        exit(1)

    in_path = argv[1]
    out_path = None
    fmt = None
    if len(argv) > 2:
        if argv[2]:
            out_path = argv[2]
        if len(argv) > 3:
            fmt = argv[3]
    else:
        out_path = None

    x = pdf_to_image(in_path, out_path, fmt)

    if isinstance(x, str):
        print(x)
    else:
        print('In memory object -> tmp.x')
        with open('tmp.x', 'wb') as tmp:
            copyfileobj(x, tmp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Conversion utilities for PostScript (PS, EPS) files to other formats.

PS and EPS files are very useful formats for creating printable figures
with `matplotlib`_. They support a number of features that are difficult
to achieve with other backends/formats, such as colored TeX strings.

This module relies on the ``[ps]`` :ref:`extra <installation-extras>`,
which implies external programs. As such, this module may be quite
OS-sensitive. Specifically, it requires the main `GhostScript`_ (GS)
program, :program:`gs`.

A small import-guarded block is provided to demo :py:func:`ps_to_image`.


.. include:: /link-defs.rst
"""

__all__ = ['gs_exe', 'ps_to_image']


import subprocess, io, shutil


#: The name of the :program:`gs` executable. Either a full path, or a
#: program that the shell can find on the :envvar:`PATH` is necessary.
gs_exe = 'gs'


#: Mappings of common MatPlotLib-like format names to sensible choices
#: among the GS output devices. Most expected names not in the map are
#: already GS output devices.
_preset_device_map = {
    'png': 'pngalpha',
    'jpg': 'jpeg',
    'bmp': 'bmp16m',
    'pdf': 'pdfwrite'
}


def ps_to_image(input_file, output_file, format='pngalpha', dpi=None):
    """
    Convert a PS or EPS document into an image file.

    EPS files are preferred inputs because they allow for proper
    trimming of the output image margins.

    This function uses the :py:mod:`subprocess` module to operate. It
    requires the presence of the :program:`gs` program from
    `GhostScript`_.

    `input_file` may be a string path or a file-like object.

    `output_file` may be a string, a file-like object or :py:obj:`None`.
    If :py:obj:`None`, an :py:class:`io.BytesIO` object containing the
    image is returned.

    `format` may be either the name of MatPlotLib-like presets or the
    name of a `GhostScript`_ output device. The following is a list of
    preset formats with the GS devices that they map to:

        - ``'png'``: ``pngalpha``
        - ``'jpg'``: ``jpeg``
        - ``'bmp'``: ``bmp16m``
        - ``'pdf'``: ``pdfwrite``

    Preset names do not overlap with any output device, so any value of
    `format` not matching a preset is interpreted as a device name. See
    the docs at http://ghostscript.com/doc/current/Devices.htm for a
    complete list of available output devices.

    `format` defaults to ``'pngalpha'``.

    Returns the name of the output file, or an in-memory file-like object
    (:py:class:`io.BytesIO`) if `output_file` is :py:obj:`None`.


    .. include:: /link-defs.rst
    """
    prog = [
        gs_exe, '-q', '-dSAFER', '-dBATCH', '-dNOPAUSE',
        '-dUseTrimBox', '-dEPSCrop'
    ]

    if format in _preset_device_map:
        device = _preset_device_map[format]
    else:
        device = format
    prog.append('-sDEVICE=' + device)

    if dpi:
        prog.append('-r{}'.format(dpi))

    rewind = False
    if isinstance(output_file, str):
        # Output to file
        prog.append('-sOutputFile=' + output_file)
        stdout = None
    else:
        prog.append('-sOutputFile=-')
        if output_file is None:
            # Output to memory
            output_file = io.BytesIO()
            rewind = True
        stdout = output_file

    if isinstance(input_file, str):
        prog.append(input_file)
        stdin = None
        input_file = None
    else:
        prog.append('-_')
        stdin = subprocess.PIPE

    proc = subprocess.Popen(prog, stdin=stdin, stdout=stdout)
    if proc.stdin:
        shutil.copyfileobj(input_file, proc.stdin)
        proc.stdin.close()
    proc.wait()

    if rewind:
        output_file.seek(0)

    return output_file


if __name__ == '__main__':
    from sys import argv

    if len(argv) <= 1:
        raise ValueError('Usage: python -m haggis.files.ps '
                         'input.[e]ps [output.* | ""] [format]')
    input_file = argv[1]
    output_file = None
    kwargs = {}
    if len(argv) > 2:
        if argv[2]:
            output_file = argv[2]
        if len(argv) > 3:
            kwargs['format'] = argv[3]
    else:
        output_file = None

    x = ps_to_image(input_file, output_file, **kwargs)

    if isinstance(x, str):
        print(x)
    else:
        print('In memory object -> tmp.x')
        with open('tmp.x', 'wb') as tmp:
            shutil.copyfileobj(x, tmp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for working with FITS files, only available when the
``[scio]`` :ref:`extra <installation-extras>` is installed.

If `astropy`_ is not found at import time, this module will have a
:py:data:`fits_enabled` attribute, which will be :py:obj:`False`. If
`astropy`_ is found, on the other hand, :py:data:`fits_enabled` will
be :py:obj:`True`, and all the dependent functions and attributes of the
module will be present.

.. py:data:: fits_enabled

   A boolean value indicating whether the ``[scio]``
   :ref:`extra <installation-extras>` has been installed. If
   :py:obj:`False`, the API will be severely limited.


.. include:: /link-defs.rst 
"""

__all__ = ['fits_enabled']


import os, tempfile, warnings


try:
    from astropy.io import fits
except ImportError:
    from .. import _display_missing_extra
    _display_missing_extra('scio', 'astropy')
    fits_enabled = False
else:
    __all__.extend(['TempFITS'])
    fits_enabled = True


if fits_enabled:
    class TempFITS():
        """
        A context manager for storing the contents of a numpy array to a
        temporary FITS file.

        If created successfuly, the file is deleted when the context
        manager exits.

        .. py:attribute:: filename

           The name of the temporary file, either passed in directly, or
           generated during initialization.

        .. py:attribute:: delete_on_close

           Indicates whether or not the underlying file will be deleted
           when the context manager exits. Default is :py:obj:`True`.

        .. py:attribute:: open_file

           Whether or not the context manager should return a file
           object opened for reading along with the filename when it
           enters. The default is to return only the filename.

        .. py:attribute:: file

           The file handle opened by when the context manager enters,
           if :py:attr:`open_file` is :py:obj:`True`. At all other
           times, this attribute is :py:obj:`None`.

        .. todo::

           Most of TempFITS can be factored out into a much more general
           base class.
        """
        def __init__(self, array, filename=None, *, open_file=True,
                     delete_on_close=True, **kwargs):
            """
            Dump the specified array to a file and record the file name.

            If `filename` is given, use it as-is and ignore all
            `kwargs`. Otherwise, pass the keyword arguments `suffix`,
            `prefix` and `dir` to :py:func:`tempfile.mkstemp` when
            creating the file. The `text` keyword may only be set to
            :py:obj:`False`. It will be ignored with a warning if
            :py:obj:`True` since FITS files are binary by their nature.

            The name of the file is stored in the :py:attr:`filename`
            attribute of this object. The named file can be closed and
            reopened at will, but keep in mind that it will be deleted
            when this context manager exits if `delete_on_close` is
            :py:obj:`True`.

            `open_file` determines if a file object open for reading is
            returned when the context manager is entered, in addition to
            the file name.
            """
            if kwargs.get('text'):
                warnings.warn('Illegal value True for "text" argument '
                              'to mkstemp for FITS files')
                kwargs['text'] = False

            if filename:
                self.filename = filename
                file = open(filename, 'wb')
            else:
                fd, self.filename = tempfile.mkstemp(**kwargs)
                file = os.fdopen(fd, 'wb')

            with file:
                fits.PrimaryHDU(array).writeto(file)

            self.delete_on_close = delete_on_close
            self.open_file = open_file
            self.file = None

        def __enter__(self):
            """
            Return the name of the file and optionally a file object
            open for reading.

            If :py:attr:`open_file` is :py:obj:`True`, a file handle
            opened for reading will be returned as a second output
            argument. The file pointer will be set to the beginning.
            """
            if self.openFile:
                self.file = open(self.filename, 'rb')
                return self.filename, self.file
            else:
                self.file = None
                return self.filename

        def __exit__(self, *exc_info):
            """
            Ensure that the file is closed regardless of any exceptions,
            and deleted if :py:attr:`delete_on_close` was chosen.

            *Any* closeable handle that is assigned to the
            :py:attr:`file` attribute will be closed, even if it was not
            the one that was opened originally, and regardless of
            :py:attr:`open_file`.
            """
            if self.file:
                self.file.close()
                self.file = None
            if self.delete_on_close:
                os.remove(self.filename)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Custom module loading functionality for Python code, wrapped around
portions of :py:mod:`importlib`.
"""

__all__ = ['load_object', 'load_module', 'module_as_dict']


from collections import deque
from importlib import import_module
from importlib.util import spec_from_loader, module_from_spec
from importlib.machinery import SourceFileLoader
from inspect import ismodule, isclass, isfunction
from os.path import basename, splitext
import sys


def load_object(name):
    """
    Imports a single object by its qualified name.

    `name` is expected to have the form ``pkg.module.object``, with all
    elements separated by dots. The last element is the name of the
    module-level attribute to load. The path is expected to be
    accessible from the normal Python path.
    """
    path = name.split('.')
    name = path[-1]
    path = '.'.join(path[:-1])
    module = import_module(path)
    return getattr(module, name)


def load_module(module, name=None, sys_module=False,
                injection_var=None, injection=None):
    """
    Load a module from a text file containing Python code.

    Parameters
    ----------
    module : str or pathlib.Path
        The path of the file to load.
    name : str or None
        The name under which the module is imported (its `__name__`
        attribute). If not supplied, or a falsy value, the name is
        computed from the file name, minus the extension. Setting this
        parameter to ``'__main__'`` will trigger import guards.
    sys_module : bool
        If truthy, add the module to :py:obj:`sys.modules` under the
        correct name. Set this to `True` if importing files that
        contain relative imports. The default is `False`.
    injection :
        Any object that the user wishes to inject into the loading
        process. The object is visible to the code of the module
        under the name given by `injection_var`.
    injection_var : str or None
        The name of an attribute to inject into the loading process.
        The `injection` object is bound to this name the in the module
        namespace. The value of `injection` is never inspected. It is
        injected or omitted based solely on the contents of this
        parameter. A falsy value (the default) skips injection.
    """
    # Import module: taken from https://stackoverflow.com/a/67692/2988730
    mod_name = name if name else splitext(basename(module))[0]
    # str(module) necessary because import machinery does't accept Paths
    mod_spec = spec_from_loader(
        mod_name, loader=SourceFileLoader(mod_name, str(module))
    )
    mod = module_from_spec(mod_spec)
    if injection_var:
        # Keyword injection based on
        # https://stackoverflow.com/a/38650878/2988730
        mod.__dict__[injection_var] = injection
    if sys_module:
        # https://stackoverflow.com/a/50395128/2988730
        sys.modules[mod_spec.name] = mod
    mod_spec.loader.exec_module(mod)
    return mod


def module_as_dict(module, name=None, *, injection=None, injection_var=None,
                   recurse_injection=True, include_var='__include_files__',
                   skip_dunder=True, skip_modules=True, skip_classes=False,
                   skip_functions=False):
    """
    Load Python module code as a dictionary.

    This function is intended to support the loading of configuration
    files that use valid Python code into a dictionary. The loaded
    module will not be inserted into :py:data:`sys.modules`.

    Basic filtering of the loaded namespace is supported: dunder
    attributes and imported modules are omitted from the final result by
    default. This behavior can be altered with the `skip_dunder` and
    `skip_modules` parameters, respectively.

    A reference can be injected into the loaded module *before* its
    code is run (i.e., making it available to the module code) using
    the `injection` and `injection_var` parameters. `injection` is the
    data itself. It is never inspected or modified in any way.
    `injection_var` names the module attribute that `injection` will
    be bound to. If the module defines a variable with the same name as
    `injection_var`, the injected reference will have no effect.

    If a loaded module contains an attribute named by the
    `include_var` parameter, it must be a sequence of paths or
    strings. All names in the sequence will be loaded recursively into
    the same dictionary as well. Includes will be loaded and parsed with
    the same parameters as the root file (except for `name` and possibly
    `injection_var`), in breadth-first order. Successive levels do *not*
    override values set by the root module that this function is called
    with. In the model for which this function was developed,
    configuration files can reference and override default static
    configurations provided externally through include files.

    Parameters
    ----------
    module : str or pathlib.Path
        The path of the module to load.
    name : str or None
        The name under which the module is imported (its `__name__`
        attribute). If not supplied, or a falsy value, the name is
        computed from the file name, minus the extension. Setting this
        parameter to ``'__main__'`` will trigger import guards.
    injection :
        Any object that the user wishes to inject into the loading
        process. The object is visible to the code of the module
        under the name given by `injection_var`.
    injection_var : str or None
        The name of an attribute to inject into the loading process.
        The `injection` object is bound to this name the in the module
        namespace. The value of `injection` is never inspected. It is
        injected or omitted based solely on the contents of this
        parameter. A falsy value (the default) skips injection.
    recurse_injection : bool
        Whether or not to provide `injection` to recursively loaded
        modules (based on `include_var`). If recursion is enabled (the
        default), the same variable name is reused.
    include_var : str or None
        The name of the attribute to look into to find additional
        include files. Defaults to ``'__include_files__'``. If Falsy,
        do not recurse.
    skip_dunders : bool
        Whether or not to skip attributes starting with a double
        underscore (``__``) when converting to a dictionary. Defaults to
        :py:obj:`True`.
    skip_modules : bool
        Whether or not to skip module objects that are found in the
        loaded namespace when converting to a dictionary. Defaults to
        :py:obj:`True`.
    skip_classes : bool
        Whether or not to skip class objects that are found in the
        loaded namespace when converting to a dictionary. Defaults to
        :py:obj:`False`.
    skip_functions : bool
        Whether or not to skip function objects that are found in the
        loaded namespace when converting to a dictionary. Defaults to
        :py:obj:`False`.
    """
    def parse_module(mod, output, submodules):
        """
        Append the contents of the module's dictionary to an existing
        dictionary of key-value pairs.

        Modules and keys beginning with double underscore (``__``) can
        be filteres. Keys are *not* overriden in the output. Duplicates
        are discarded from the module being loaded and not overwritten.

        Parameters
        ----------
        mod : ModuleType
            The module to parse.
        output : dict
            The mutable dictionary to insert data into, constructed
            externally.
        submodules : deque
            A sequence of modules that are loaded as they are found in
            `include_var`. These modules will be parsed in the order
            found, maintaining the breadth-first contract.
        """
        # First extract the modules dictionary:
        mod = vars(mod)
        # Get the includes (so `skip_dunders` doesn't filter `include_var`)
        if include_var:
            inject = injection_spec if recurse_injection else {}
            submodules.extend(load_module(name, None, **inject)
                              for name in mod.get(include_var, []))

        # Create a filter method
        def filter_func(d, k, v):
            if k in d or (skip_dunder and k.startswith('__')):
                return False
            if ((skip_modules and ismodule(v)) or
                (skip_classes and isclass(v)) or
                (skip_functions and isfunction(v))):
                return False
            return True

        # Update the dictionary: use a generator for speed and efficiency
        dictionary.update(item for item in mod.items()
                          if filter_func(dictionary, *item))

    # Convert the module to a dictionary
    dictionary = {}
    submodules = deque()
    injection_spec = {'injection_var': injection_var, 'injection': injection}
    # Load the root module
    submodules.append(load_module(module, name, **injection_spec))

    # Keep adding elements to the dictionary as long as there are includes
    while submodules:
        parse_module(submodules.popleft(), dictionary, submodules)

    # Return the dictionary
    return dictionary
